"""A meta variable definition that is used for passing information along with a sensor graph."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement


@python_2_unicode_compatible
class MetaStatement(SensorGraphStatement):
    """A metadata variable definition.

    The statement form is:
    meta var = value;

    where var must be an identifier and value must be an rvalue.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
    """

    def __init__(self, parsed):
        self.identifier = parsed[0]
        self.value = parsed[1]

        super(MetaStatement, self).__init__([])

    def __str__(self):
        return u'meta %s = %s;' % (str(self.identifier), str(self.value))



from .scope import Scope
from ...node import InputTrigger
from ...exceptions import SensorGraphSemanticError
from ...stream import DataStream

class TriggerScope(Scope):
    """A scope that implements trigger_chain but cannot provide a clock.

    Args:
        sensor_graph (SensorGraph): The sensor graph we are working on.
        scope_stack (list(Scope)): The stack of already allocated scopes.
        trigger_input ((DataStream, InputTrigger): The input trigger stream
            and condition.  The input trigger must already be attached.
    """

    def __init__(self, sensor_graph, scope_stack, trigger_input):
        parent = scope_stack[-1]
        alloc = parent.allocator
        sensor_graph = parent.sensor_graph

        super(TriggerScope, self).__init__(u"Trigger Scope", sensor_graph, alloc, parent)

        # Create our own node to create our triggering chain
        # this will be optimized out if it turned out not to be needed
        # after all nodes have been allocated.  Since we can trigger on a root
        # node, make sure we don't try to make our output a root input.
        stream_type = trigger_input[0].stream_type
        if stream_type == DataStream.InputType:
            stream_type = DataStream.UnbufferedType

        stream = alloc.allocate_stream(stream_type)
        sensor_graph.add_node(u'({} {}) => {} using copy_latest_a'.format(trigger_input[0], trigger_input[1], stream))
        self.trigger_stream = stream
        self.trigger_cond = InputTrigger(u'count', '==', 1)

    def trigger_chain(self):
        """Return a NodeInput tuple for creating a node.

        Returns:
            (StreamIdentifier, InputTrigger)
        """

        trigger_stream = self.allocator.attach_stream(self.trigger_stream)
        return (trigger_stream, self.trigger_cond)

    def clock(self, interval):
        """Return a NodeInput tuple for triggering an event every interval.

        Args:
            interval (int): The interval (in seconds) at which this input should
                trigger.
        """

        raise SensorGraphSemanticError("TriggerScope does not provide a clock output that can be used.")



"""Time based event block for scheduling RPCs every X interval."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ..scopes import TriggerScope

@python_2_unicode_compatible
class EveryBlock(SensorGraphStatement):
    """A block of statements that should run every time interval

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        children(list(SensorGraphStatement)): The statements that are
            part of this every block.
    """

    def __init__(self, parsed, children):
        self.interval = parsed[0]

        super(EveryBlock, self).__init__(children)

    def __str__(self):
        return u"every %s" % (str(self.interval),)

    def execute_before(self, sensor_graph, scope_stack):
        """Execute statement before children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        new_scope = TriggerScope(sensor_graph, scope_stack, parent.clock(self.interval))
        scope_stack.append(new_scope)

    def execute_after(self, sensor_graph, scope_stack):
        """Execute statement after children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        scope_stack.pop()



from .meta_statement import MetaStatement
from .config_statement import SetConfigStatement
from .config_block import ConfigBlock
from .call_statement import CallRPCStatement
from .copy_statement import CopyStatement
from .every_block import EveryBlock
from .on_block import OnBlock
from .when_block import WhenBlock
from .latch_block import LatchBlock
from .streamer_statement import StreamerStatement
from .trigger_statement import TriggerStatement


# All of the known statements in a sensor graph file and their matching tag names
statement_map = {
    'meta_statement': MetaStatement,
    'set_statement': SetConfigStatement,
    'config_block': ConfigBlock,
    'call_statement': CallRPCStatement,
    'copy_statement': CopyStatement,
    'streamer_statement': StreamerStatement,
    'trigger_statement': TriggerStatement,
    'every_block': EveryBlock,
    'on_block': OnBlock,
    'when_block': WhenBlock,
    'latch_block': LatchBlock
}

__all__ = ['statement_map']



from builtins import str
import pytest

from iotile.sg.node_descriptor import parse_node_descriptor
from iotile.sg.model import DeviceModel
from iotile.sg import SensorGraph, SensorLog, DataStream

def test_basic_parsing():
    """Make sure we can parse a basic node description."""

    node, inputs, processing = parse_node_descriptor('(input 1 always && input 2 when count >= 1) => buffered node 1 using copy_all_a', DeviceModel())

    assert processing == u'copy_all_a'
    assert str(node.stream) == u'buffered 1'
    assert str(inputs[0][0]) == u'input 1'
    assert inputs[0][1] is None

    assert str(inputs[1][0]) == u'input 2'
    assert inputs[1][1] is not None

def test_string_generation():
    """Make sure we can print nodes."""

    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    sg.add_node('(input 1 always && input 2 when count >= 1) => buffered node 1 using copy_all_a')
    assert str(sg.nodes[-1]) == u'(input 1 always && input 2 when count >= 1) => buffered 1 using copy_all_a'

    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)
    sg.add_node('(input 1 when value < 0x10) => buffered node 1 using copy_all_a')
    assert str(sg.nodes[-1]) == u'(input 1 when value < 16) => buffered 1 using copy_all_a'



"""An in memory storage engine for sensor graph."""

from builtins import str
from iotile.sg import DataStream
from iotile.sg.exceptions import StorageFullError, StreamEmptyError


class InMemoryStorageEngine(object):
    """A simple in memory storage engine for sensor graph.

    Args:
        model (DeviceModel): A model for the device type that we are
            emulating so that we can constrain our total memory
            size appropriately to get the same behavior that would
            be seen on an actual device.
    """

    def __init__(self, model):
        self.model = model
        self.storage_length = model.get(u'max_storage_buffer')
        self.streaming_length = model.get(u'max_streaming_buffer')
        self.streaming_data = []
        self.storage_data = []

    def count(self):
        """Count the number of readings.

        Args:
            stream (DataStream): The stream to count readings in

        Returns:
            (int, int): The number of readings in storage and streaming
        """

        return (len(self.storage_data), len(self.streaming_data))

    def clear(self):
        """Clear all data from this storage engine."""

        self.storage_data = []
        self.streaming_data = []

    def push(self, value):
        """Store a new value for the given stream.

        Args:
            value (IOTileReading): The value to store.  The stream
                parameter must have the correct value
        """

        stream = DataStream.FromEncoded(value.stream)

        if stream.stream_type == DataStream.OutputType:
            if len(self.streaming_data) == self.streaming_length:
                raise StorageFullError('Streaming buffer full')

            self.streaming_data.append(value)
        else:
            if len(self.storage_data) == self.storage_length:
                raise StorageFullError('Storage buffer full')

            self.storage_data.append(value)

    def get(self, buffer_type, offset):
        """Get a reading from the buffer at offset.

        Offset is specified relative to the start of the data buffer.
        This means that if the buffer rolls over, the offset for a given
        item will appear to change.  Anyone holding an offset outside of this
        engine object will need to be notified when rollovers happen (i.e.
        popn is called so that they can update their offset indices)

        Args:
            buffer_type (str): The buffer to pop from (either u"storage" or u"streaming")
            offset (int): The offset of the reading to get
        """

        if buffer_type == u'streaming':
            chosen_buffer = self.streaming_data
        else:
            chosen_buffer = self.storage_data

        if offset >= len(chosen_buffer):
            raise StreamEmptyError("Invalid index given in get command", requested=offset, stored=len(chosen_buffer), buffer=buffer_type)

        return chosen_buffer[offset]

    def popn(self, buffer_type, count):
        """Remove and return the oldest count values from the named buffer

        Args:
            buffer_type (str): The buffer to pop from (either u"storage" or u"streaming")
            count (int): The number of readings to pop

        Returns:
            list(IOTileReading): The values popped from the buffer
        """

        buffer_type = str(buffer_type)

        if buffer_type == u'streaming':
            chosen_buffer = self.streaming_data
        else:
            chosen_buffer = self.storage_data

        if count > len(chosen_buffer):
            raise StreamEmptyError("Not enough data in buffer for popn command", requested=count, stored=len(chosen_buffer), buffer=buffer_type)

        popped = chosen_buffer[:count]
        remaining = chosen_buffer[count:]

        if buffer_type == u'streaming':
            self.streaming_data = remaining
        else:
            self.storage_data = remaining

        return popped




from iotile.core.exceptions import IOTileException


class StreamEmptyError(IOTileException):
    """Thrown when a stream walker is empty."""

    pass


class StorageFullError(IOTileException):
    """Thrown when more readings can not be stored because the storage area is full."""

    pass


class TooManyOutputsError(IOTileException):
    """Thrown when there are more outputs required for a node than it can support."""

    pass


class TooManyInputsError(IOTileException):
    """Thrown when there are more inputs to a node than its maximum number of inputs."""

    pass


class NodeConnectionError(IOTileException):
    """Thrown when a node is created with its inputs unconnected.

    Nodes that depend on the output of other nodes need to be created after
    those nodes have already been created.  If a node is created and its inputs
    do not refer either to global inputs or to other nodes that have already
    been created, this exception is thrown.
    """

    pass


class ProcessingFunctionError(IOTileException):
    """Thrown when a processing function for a node is invalid.

    This can either be thrown during an attempt to execute the processing
    function if it cannot be found or was not set correctly.  It can also
    be thrown during serializing a sensor graph for embedding into an IOTile
    Device if the procesing function is not available on the embedded device.
    """

    pass


class SensorGraphSyntaxError(IOTileException):
    """Thrown when there is an error parsing a sensor graph description file.

    The parameters line and column will indicate where the error occured and
    other paremeters will give more specific information about what was expected.
    """

    pass


class SensorGraphSemanticError(IOTileException):
    """Thrown when a syntactically valid statement occurs in a context that is not valid.

    Not all statements can appear in all scopes.  This exception indicates a semantic error
    where a statement appears in a scope that does not allow it.
    """

    pass


class UnresolvedIdentifierError(IOTileException):
    """Thrown when an identifier is asked for but cannot be found.

    The name of the identifier is given in the parameters.  This exception will
    also be thrown if an identifier is found but resolves to an object of an incorrect
    type.  For example, if you are asking for a DataStream but the identifier names
    a SlotIdentifier.
    """

    pass



from ...node import InputTrigger
from ...stream import DataStream
from ...exceptions import SensorGraphSemanticError
from .scope import Scope


class ClockScope(Scope):
    """A scope that implements trigger_chain and can provide a clock.

    Args:
        sensor_graph (SensorGraph): The sensor graph we are working on.
        scope_stack (list(Scope)): The stack of already allocated scopes.
        clock_input ((DataStream, InputTrigger)): The input clock stream
            and trigger conditions used to create this clock.
        interval (int): The interval at which the input clock stream gets
            updated in seconds.
    """

    def __init__(self, sensor_graph, scope_stack, clock_input, interval):
        parent = scope_stack[-1]
        alloc = parent.allocator
        sensor_graph = parent.sensor_graph

        super(ClockScope, self).__init__(u"Sub-Clock Scope", sensor_graph, alloc, parent)

        # Create our own node to create our clock chain
        # this will be optimized out if it turned out not to be needed
        # after all nodes have been allocated
        stream = alloc.allocate_stream(DataStream.CounterType)
        sensor_graph.add_node(u'({} {}) => {} using copy_all_a'.format(clock_input[0], clock_input[1], stream))

        self.clock_stream = stream
        self.clock_interval = interval

    def clock(self, interval):
        """Return a NodeInput tuple for triggering an event every interval.

        Args:
            interval (int): The interval (in seconds) at which this input should
                trigger.
        """

        if (interval % self.clock_interval) != 0:
            raise SensorGraphSemanticError("ClockScope was asked for a clock that was not a multiple of its internal clock", internal_interval=self.clock_interval, requested_interval=interval)

        stream = self.allocator.attach_stream(self.clock_stream)

        return (stream, InputTrigger(u'count', '==', interval // self.clock_interval))




import os
import pytest
from iotile.sg.exceptions import SensorGraphSyntaxError
from iotile.sg import DataStream, SlotIdentifier, DataStreamSelector
from iotile.sg.parser import SensorGraphFileParser
import iotile.sg.parser.language as language


@pytest.fixture
def parser():
    return SensorGraphFileParser()


def get_path(name):
    return os.path.join(os.path.dirname(__file__), 'sensor_graphs', name)


def test_language_constructs():
    """Make sure the basic sensor graph language constructs work."""

    # Create the parser
    language.get_language()

    # Test time interval parsing
    parsed = language.time_interval.parseString('1 day')
    assert parsed.asList()[0] == 60*60*24

    # Test block id parsing
    parsed = language.block_id.parseString('every 1 day')
    assert parsed.asList()[0][0] == 60*60*24

    # Test block parsing
    parsed = language.block_bnf.parseString('every 1 day {}')
    assert parsed.asList()[0][0][0] == 60*60*24

    # Test stream parsing
    parsed = language.stream.parseString('output 1')
    assert isinstance(parsed.asList()[0], DataStream)

    # Test call_rpc statement
    parsed = language.callrpc_stmt.parseString(u'call 0x4001 on slot 1 => output 1;')
    assert parsed.asList()[0][0] == 0x4001

    # Test block with statement parsing
    parsed = language.block_bnf.parseString(u'every 1 day { call 0x5001 on slot 2 => output 1; }')
    assert parsed.asList()[0][0][0] == 60*60*24

    # Test parsing stream_trigger
    parsed = language.stream_trigger.parseString(u'value(input 2) == 10')
    assert parsed[0].getName() == u'stream_trigger'

    parsed = language.stream_trigger.parseString(u'count(output 1) <= 10')
    assert parsed[0].getName() == u'stream_trigger'

    # Test parsing on block with identifier
    parsed = language.block_bnf.parseString(u'on test_identifier {}')
    assert parsed[0][0][0][0].getName() == u'identifier'
    assert parsed[0][0][0][0][0] == u'test_identifier'

    parsed = language.block_bnf.parseString(u'on value(input 2) >= 5 {}')
    assert parsed[0][0][0][0].getName() == u'stream_trigger'
    assert parsed[0][0][0][0][0] == u'value'
    assert parsed[0][0][0][0][1] == DataStream.FromString('input 2')
    assert parsed[0][0][0][0][2] == u'>='
    assert parsed[0][0][0][0][3] == 5

    # Test parsing on block with 2 conditions
    parsed = language.block_bnf.parseString(u'on test_identifier and hello_id {}')
    assert parsed[0][0][0][0].getName() == u'identifier'
    assert parsed[0][0][0][0][0] == u'test_identifier'
    assert parsed[0][0][2][0].getName() == u'identifier'
    assert parsed[0][0][2][0][0] == u'hello_id'
    assert parsed[0][0][1] == u'and'

    parsed = language.block_bnf.parseString(u'on test_identifier or value(constant 1) == 2 {}')
    assert parsed[0][0][0][0].getName() == u'identifier'
    assert parsed[0][0][0][0][0] == u'test_identifier'
    assert parsed[0][0][1] == u'or'

    # Test parser streamer statements
    parsed = language.streamer_stmt.parseString(u'manual streamer on output 1;')
    assert parsed[0]['selector'] == DataStreamSelector.FromString('output 1')

    parsed = language.streamer_stmt.parseString(u'manual streamer on all system outputs;')
    assert parsed[0]['selector'] == DataStreamSelector.FromString('all system outputs')

    parsed = language.streamer_stmt.parseString(u'manual signed streamer on output 1 to slot 1;')
    assert parsed[0]['selector'] == DataStreamSelector.FromString('output 1')
    assert parsed[0]['explicit_tile'] == SlotIdentifier.FromString('slot 1')
    assert parsed[0]['security'] == u'signed'

    # Test parsing copy statements
    parsed = language.simple_statement.parseString(u'copy unbuffered 1 => unbuffered 2;')
    assert parsed[0]['explicit_input'][0] == DataStream.FromString('unbuffered 1')

    parsed = language.simple_statement.parseString(u'copy 15 => unbuffered 2;')
    assert parsed[0]['constant_input'] == 15

    parsed = language.simple_statement.parseString(u'copy 0x20 => unbuffered 2;')
    assert parsed[0]['constant_input'] == 0x20


def test_basic_parsing(parser):
    """Make sure we can parse a basic file without syntax errors."""

    parser.parse_file(get_path('basic_meta_file.sgf'))


def test_every_interval_block_parsing(parser):
    """Make sure we can parse an interval block."""

    parser.parse_file(get_path('basic_block.sgf'))


def test_statement_syntax_error(parser):
    """Make sure we get a nice syntax error when we type a bad statement."""

    with pytest.raises(SensorGraphSyntaxError) as exc_info:
        parser.parse_file(get_path('syntax_error_statement.sgf'))

    assert exc_info.value.params['line_number'] == 10


def test_nested_blocks(parser):
    """Make sure we can parse nested blocks."""

    parser.parse_file(get_path('nested_block.sgf'))


def test_ignoring_comments(parser):
    """Make sure we ignore comments."""

    parser.parse_file(get_path('comment_graph.sgf'))



from .simulator import SensorGraphSimulator

# FIXME: add this back once we merge the port of py36 compatible typedargs
# from .hosted_executor import SemihostedRPCExecutor

__all__ = ['SensorGraphSimulator']#, 'SemihostedRPCExecutor']



"""Optimizers that can reduce the resources required to run a sensor graph.

Optimizers can eliminate redundant nodes that are provably not needed because
their results are known in advance and have no side effects.  They can also
remove dead code that can never be triggered.
"""

from .optimizer import SensorGraphOptimizer


__all__ = ['SensorGraphOptimizer']



"""Command line script to compile a sensor graph."""

import sys
import argparse
from builtins import str
from iotile.sg import SensorGraph, SensorLog, DeviceModel
from iotile.sg.parser import SensorGraphFileParser
from iotile.sg.optimizer import SensorGraphOptimizer
from iotile.sg.output_formats import known_formats


def build_args():
    """Create command line argument parser."""

    parser = argparse.ArgumentParser(description=u'Compile a sensor graph.')
    parser.add_argument(u'sensor_graph', type=str, help=u"the sensor graph file to load and run.")
    parser.add_argument(u'-f', u'--format', default=u"nodes", choices=[u'nodes', u'ast', 'snippet'], type=str, help=u"the output format for the compiled result.")
    parser.add_argument(u'-o', u'--output', type=str, help=u"the output file to save the results (defaults to stdout)")
    parser.add_argument(u'--disable-optimizer', action="store_true", help=u"disable the sensor graph optimizer completely")
    return parser


def main():
    arg_parser = build_args()
    args = arg_parser.parse_args()

    model = DeviceModel()

    parser = SensorGraphFileParser()
    parser.parse_file(args.sensor_graph)

    outfile = sys.stdout

    if args.output is not None:
        outfile = open(args.output, "wb")

    if args.format == u'ast':
        outfile.write(parser.dump_tree())
        outfile.close()
        sys.exit(0)

    parser.compile(model)

    if not args.disable_optimizer:
        opt = SensorGraphOptimizer()
        opt.optimize(parser.sensor_graph, model=model)

    if args.format == u'nodes':
        for node in parser.sensor_graph.dump_nodes():
            outfile.write(node + u'\n')
    else:
        if args.format not in known_formats:
            print("Unknown output format: {}".format(args.format))
            outfile.close()
            sys.exit(1)

        output = known_formats[args.format](parser.sensor_graph)
        outfile.write(output)

    outfile.close()



"""Scopes provide the vessel for storing key state data during code generation."""

from builtins import str
from iotile.sg.exceptions import UnresolvedIdentifierError, SensorGraphSemanticError


class Scope(object):
    """An environment used during code generation.

    Scopes contain all of the state information used
    to assign triggers to nodes and link them into a
    SensorGraph during the code generation process.

    They have four main entry points.  In what follows,
    NodeInput refers to a tuple of StreamIdentifier and
    InputTrigger that can be used to connect a node to
    a Streamwalker with a trigger.

    - trigger_chain(): Allocate a NodeInput that will
      trigger a node when the previous node in this
      chain is finished.
    - clock(interval): Allocate a NodeInput that will
      trigger every interval seconds.
    - resolve_event(event_name): Resolve an event_name
      to a NodeInput.
    - resolve_identifier(identifier_name): Resolve an
      identifier to an object

    Args:
        name (str): An identifier for this scope to be
            used in debug messages
        sensor_graph (SensorGraph): The SensorGraph we
            are operating on
        allocator (StreamAllocator): The global stream
            allocator we are using
        parent (Scope): Our parent scope if we have one
            so that we can forward on requests for
            information if needed.
    """

    def __init__(self, name, sensor_graph, allocator, parent):
        self.sensor_graph = sensor_graph
        self.parent = parent
        self.name = name
        self.allocator = allocator
        self._known_identifiers = {}

    def add_identifier(self, name, obj):
        """Add a known identifier resolution.

        Args:
            name (str): The name of the identifier
            obj (object): The object that is should resolve to
        """

        name = str(name)
        self._known_identifiers[name] = obj

    def resolve_identifier(self, name, expected_type=None):
        """Resolve an identifier to an object.

        There is a single namespace for identifiers so the user also should
        pass an expected type that will be checked against what the identifier
        actually resolves to so that there are no surprises.

        Args:
            name (str): The name that we want to resolve
            expected_type (type): The type of object that we expect to receive.
                This is an optional parameter.  If None is passed, no type checking
                is performed.

        Returns:
            object: The resolved object
        """

        name = str(name)

        if name in self._known_identifiers:
            obj = self._known_identifiers[name]
            if expected_type is not None and not isinstance(obj, expected_type):
                raise UnresolvedIdentifierError(u"Identifier resolved to an object of an unexpected type", name=name, expected_type=expected_type.__name__, resolved_type=obj.__class__.__name__)

            return obj

        if self.parent is not None:
            try:
                return self.parent.resolve_identifier(name)
            except UnresolvedIdentifierError:
                pass

        raise UnresolvedIdentifierError(u"Could not resolve identifier", name=name, scope=self.name)

    def trigger_chain(self):
        """Return a NodeInput tuple for creating a node.

        Returns:
            (StreamIdentifier, InputTrigger)
        """

        raise SensorGraphSemanticError("There is no trigger chain in this scope since no triggering criteria have been set", scope=self.name)

    def clock(self, interval):
        """Return a NodeInput tuple for triggering an event every interval.

        Args:
            interval (int): The interval (in seconds) at which this input should
                trigger.
        """

        raise SensorGraphSemanticError("There is not default clock defined in this scope", scope=self.name)


import pytest
from iotile.sg.sim import SensorGraphSimulator
from iotile.sg.slot import SlotIdentifier
from iotile.sg.known_constants import config_user_tick_secs
from iotile.sg import DeviceModel, SensorLog, SensorGraph, DataStream
from iotile.core.hw.reports import IOTileReading

@pytest.fixture
def basic_sg():
    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    sg.add_node('(system input 2 always) => unbuffered 1 using copy_all_a')

    return sg


@pytest.fixture
def usertick_sg():
    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    sg.add_node('(system input 3 always) => counter 1 using copy_latest_a')
    sg.add_config(SlotIdentifier.FromString('controller'), config_user_tick_secs, 'uint32_t', 2)

    return sg


@pytest.fixture
def callrpc_sg():
    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    sg.add_node('(system input 2 always && constant 1 always) => unbuffered 2 using call_rpc')
    log.push(DataStream.FromString('constant 1'), IOTileReading(0, 0, 0x000a8000))

    return sg

def test_basic_sim(basic_sg):
    """Make sure we can run a very simple sensor graph and have it work."""

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 1000 seconds')
    sim.run(basic_sg)

    # Make sure the sensor graph ran correctly
    last_input = basic_sg.sensor_log.inspect_last(DataStream.FromString('system input 2'))
    last_output = basic_sg.sensor_log.inspect_last(DataStream.FromString('unbuffered 1'))

    assert last_input.value == 1000
    assert last_output.value == 1000
    assert sim.tick_count == 1000


def test_rpc_sim(callrpc_sg):
    """Make sure we can run a very simple sensor graph and have it work."""

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 1000 seconds')
    sim.run(callrpc_sg)

    # Make sure the sensor graph ran correctly
    last_input = callrpc_sg.sensor_log.inspect_last(DataStream.FromString('system input 2'))
    last_output = callrpc_sg.sensor_log.inspect_last(DataStream.FromString('unbuffered 2'))

    assert last_input.value == 1000
    assert last_output.value == 0


def test_multiple_run_calls(callrpc_sg):
    """Make sure we can call run multiple times."""

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 100 seconds')

    sim.run(callrpc_sg)

    # Make sure the sensor graph ran correctly
    last_input = callrpc_sg.sensor_log.inspect_last(DataStream.FromString('system input 2'))
    last_output = callrpc_sg.sensor_log.inspect_last(DataStream.FromString('unbuffered 2'))

    assert last_input.value == 100
    assert last_output.value == 0

    sim.run(callrpc_sg)

    # Make sure the sensor graph ran correctly
    last_input = callrpc_sg.sensor_log.inspect_last(DataStream.FromString('system input 2'))
    last_output = callrpc_sg.sensor_log.inspect_last(DataStream.FromString('unbuffered 2'))

    assert last_input.value == 200
    assert last_output.value == 0


def test_usertick(usertick_sg):
    """Make sure we receive user ticks in the simulation."""

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 100 seconds')

    sim.run(usertick_sg)

    # Make sure the sensor graph ran correctly
    last_input = usertick_sg.sensor_log.inspect_last(DataStream.FromString('system input 3'))
    last_output = usertick_sg.sensor_log.inspect_last(DataStream.FromString('counter 1'))

    assert last_input.value == 100
    assert last_output.value == 100

    sim.run(usertick_sg)

    # Make sure the sensor graph ran correctly
    last_input = usertick_sg.sensor_log.inspect_last(DataStream.FromString('system input 3'))
    last_output = usertick_sg.sensor_log.inspect_last(DataStream.FromString('counter 1'))

    assert last_input.value == 200
    assert last_output.value == 200



"""Slot identifier parsing of 'slot 1' into SlotIdentifier."""

from builtins import str
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError


@python_2_unicode_compatible
class SlotIdentifier(object):
    """A slot identifier specifies the address of a tile on TileBus.

    Slots identifiers can be built from strings using FromString().
    The form of the string should be:

    slot <number> where number is between 1 and 31, inclusive
    or
    controller

    which indicates that this is a controller tile (implicitly slot 0)

    Args:
        slot (int): The slot number bertween 1 and 32 if this tile is not
            a controller.  The controller argument must be False if this
            argument is specified.
        controller (bool): True if this is a controller tile, in which case
            the slot number must be None.
    """

    def __init__(self, slot=None, controller=False):

        if slot is not None:
            slot = int(slot)

        if controller and (slot is not None):
            raise ArgumentError("You cannot pass both a slot and controller argument to SlotIdentifier", slot=slot, controller=controller)

        if slot is not None and (slot < 1 or slot >= 32):
            raise ArgumentError("Slot number too big.  It must be between 1 and 31, inclusive", slot=slot)

        self.controller = controller
        self.slot = slot

    @property
    def address(self):
        """The address of this tile.

        Tile addresses are calculated as:
        controller: 8
        other tiles: 9 + slot (so slot 1 is 10)
        """

        if self.controller:
            return 8

        return 10 + self.slot

    @classmethod
    def FromString(cls, desc):
        """Create a slot identifier from a string description.

        The string needs to be either:

        controller
        OR
        slot <X> where X is an integer that can be converted with int(X, 0)

        Args:
            desc (str): The string description of the slot

        Returns:
            SlotIdentifier
        """

        desc = str(desc)

        if desc == u'controller':
            return SlotIdentifier(controller=True)

        words = desc.split()
        if len(words) != 2 or words[0] != u'slot':
            raise ArgumentError(u"Illegal slot identifier", descriptor=desc)

        try:
            slot_id = int(words[1], 0)
        except ValueError:
            raise ArgumentError(u"Could not convert slot identifier to number", descriptor=desc, number=words[1])

        return SlotIdentifier(slot=slot_id)

    def __str__(self):
        if self.controller:
            return u'controller'

        return u'slot {}'.format(self.slot)

    def __hash__(self):
        return hash((self.controller, self.slot))

    def __eq__(self, other):
        if not isinstance(other, SlotIdentifier):
            return NotImplemented

        return self.controller == other.controller and self.slot == other.slot



from .stream import DataStream, DataStreamSelector
from .exceptions import StreamEmptyError
from .graph import SensorGraph
from .model import DeviceModel
from .sensor_log import SensorLog
from .slot import SlotIdentifier

__all__ = ['DeviceModel', 'DataStream', 'SensorGraph', 'DataStreamSelector',
           'StreamEmptyError',  'SensorLog', 'SlotIdentifier']




from .remove_copylatest import RemoveCopyLatestPass
from .convert_to_always import ConvertCountOneToAlways
from .downgrade_copyall import ConvertCopyAllToCopyLatest

__all__ = ['RemoveCopyLatestPass', 'ConvertCountOneToAlways', 'ConvertCopyAllToCopyLatest']



"""All of the known stop conditions that we support."""

from pyparsing import Word, Regex, nums, hexnums, Literal, Optional, Group, oneOf, QuotedString, ParseException
from iotile.core.exceptions import ArgumentError

hex_number = (Regex(u'0x[0-9a-fA-F]+')).setParseAction(lambda s, l, t: [int(t[0], 0)])
dec_number = Word(nums).setParseAction(lambda s, l, t: [int(t[0], 10)])
number = hex_number | dec_number

time_unit_multipliers = {
    u'second': 1,
    u'seconds': 1,
    u'minute': 60,
    u'minutes': 60,
    u'hour': 60*60,
    u'hours': 60*60,
    u'day': 60*60*24,
    u'days': 60*60*24,
    u'month': 60*60*24*30,
    u'months': 60*60*24*30,
    u'year': 60*60*24*365,
    u'years': 60*60*24*365,
}
time_unit = oneOf(u"second seconds minute minutes hour hours day days week weeks month months year years")
time_interval = (number + time_unit).setParseAction(lambda s, l, t: [t[0]*time_unit_multipliers[t[1]]])


class StopCondition(object):
    """A condition under which the simulation should stop.

    Subclasses should override the one public method
    named should_stop(self, abs_seconds, rel_seconds, sensor_graph).

    The abs_seconds parameter is the number of seconds that have expired
    since the start of the simulation (over all calls to run).  The
    rel_seconds parameter counts relative seconds inside the current call to run.
    The sensor_graph is the sensor_graph (including data) being simulated.  The function
    should return a bool with True meaning that the simulation
    should stop.

    There should be a second class method, FromString(cls, desc) that
    tries to parse this stop condition from a text string.  The function
    must raise an ArgumentError if it could not match the input string.
    """

    def should_stop(self, abs_second_count, rel_second_count, sensor_graph):
        """Check if this stop condition is fulfilled.

        Args:
            abs_second_count (int): The number of seconds that
                have expired since the start of the simulation.
            second_count (int): The number of seconds that
                have expired since the start of the last `run` calls.
            sensor_graph (SensorGraph): The sensor graph that is
                being simulated, giving access to its stored data.

        Returns:
            bool: True if we should stop, otherwise False
        """

        return False


class TimeBasedStopCondition(StopCondition):
    """Stop the simulation after a fixed period of time.

    This time is relative to the call to `run` so the simulation
    may be continued over multiple run calls, each of which lasts
    for max_time seconds.

    Args:
        max_time (int): The maximum number of seconds to run the
            simulation for.
    """

    def __init__(self, max_time):
        self.max_time = max_time

    def should_stop(self, abs_seconds, rel_seconds, sensor_graph):
        """Check if this stop condition is fulfilled.

        Args:
            second_count (int): The number of seconds that
                have expired since the start of the simulation.
            sensor_graph (SensorGraph): The sensor graph that is
                being simulated, giving access to its stored data.

        Returns:
            bool: True if we should stop, otherwise False
        """

        return rel_seconds >= self.max_time

    @classmethod
    def FromString(cls, desc):
        """Parse this stop condition from a string representation.

        The string needs to match:
        run_time number [seconds|minutes|hours|days|months|years]

        Args:
            desc (str): The description

        Returns:
            TimeBasedStopCondition
        """

        parse_exp = Literal(u'run_time').suppress() + time_interval(u'interval')

        try:
            data = parse_exp.parseString(desc)
            return TimeBasedStopCondition(data[u'interval'][0])
        except ParseException:
            raise ArgumentError(u"Could not parse time based stop condition")



"""Set config statement that adds a config variable to the sensor graph."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from iotile.sg import SlotIdentifier
from iotile.sg.exceptions import SensorGraphSemanticError, UnresolvedIdentifierError


@python_2_unicode_compatible
class SetConfigStatement(SensorGraphStatement):
    """A config variable assignment.

    The statement form is:
    set var = value [as type];

    where var must be an identifier or a number and value must be an rvalue.
    If var is not an identifier that corresponds with a known ConfigDefinition
    that specifies the type of the config variable, the type must be specified
    explicitly.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
    """

    def __init__(self, parsed):
        self.identifier = parsed[0]
        self.value = parsed[1]
        self.explicit_type = None

        if len(parsed) == 3:
            self.explicit_type = parsed[2]

        super(SetConfigStatement, self).__init__([])

    def __str__(self):
        if self.explicit_type is not None:
            return u"set %s = %s as %s;" % (str(self.identifier), str(self.value), str(self.explicit_type))

        return u"set %s = %s" % (str(self.identifier), str(self.value))

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This adds a single config variable assignment to the current sensor graph

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]

        try:
            slot = parent.resolve_identifier('current_slot', SlotIdentifier)
        except UnresolvedIdentifierError:
            raise SensorGraphSemanticError("set config statement used outside of config block")

        if self.explicit_type is None or not isinstance(self.identifier, int):
            raise SensorGraphSemanticError("Config variable type definitions are not yet supported")

        sensor_graph.add_config(slot, self.identifier, self.explicit_type, self.value)



import pdb

import os
import pytest
from iotile.sg.exceptions import SensorGraphSyntaxError, StreamEmptyError
from iotile.sg import DataStream, DeviceModel, DataStreamSelector, SlotIdentifier
from iotile.sg.parser import SensorGraphFileParser
from iotile.sg.sim import SensorGraphSimulator
from iotile.sg.known_constants import user_connected
import iotile.sg.parser.language as language
from iotile.core.hw.reports import IOTileReading


@pytest.fixture
def parser():
    return SensorGraphFileParser()


def get_path(name):
    return os.path.join(os.path.dirname(__file__), 'sensor_graphs', name)


def test_every_block_compilation(parser):
    """Make sure we can compile a simple every block."""

    parser.parse_file(get_path(u'basic_every_1min.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    assert len(sg.nodes) == 5

    sg.load_constants()
    # Now make sure it produces the right output
    counter15 = log.create_walker(DataStreamSelector.FromString('counter 15'))
    counter16 = log.create_walker(DataStreamSelector.FromString('counter 16'))

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 120 seconds')
    sim.run(sg)

    assert counter15.count() == 2
    assert counter16.count() == 2


def test_every_block_with_buffering(parser):
    """Make sure we can compile and simulate an every block with buffered data."""

    parser.parse_file(get_path(u'basic_output.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    assert len(sg.nodes) == 5

    sg.load_constants()
    # Now make sure it produces the right output
    output1 = log.create_walker(DataStreamSelector.FromString('output 1'))
    buffered1 = log.create_walker(DataStreamSelector.FromString('buffered 1'))

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 120 seconds')
    sim.run(sg)

    assert output1.count() == 12
    assert buffered1.count() == 12


def test_streamers(parser):
    """Make sure we can compile streamer statements."""

    parser.parse_file(get_path(u'basic_streamer.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    streamers = parser.sensor_graph.streamers
    assert len(streamers) == 2
    assert streamers[0].dest == SlotIdentifier.FromString('controller')
    assert streamers[1].dest == SlotIdentifier.FromString('slot 1')
    assert streamers[1].with_other == 0



def test_every_block_splitting(parser):
    """Make sure we can split nodes in an every block."""

    parser.parse_file(get_path(u'basic_every_split.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    assert len(sg.nodes) == 8

    # Now make sure it produces the right output
    counter15 = log.create_walker(DataStreamSelector.FromString('counter 15'))
    counter16 = log.create_walker(DataStreamSelector.FromString('counter 16'))
    counter17 = log.create_walker(DataStreamSelector.FromString('counter 16'))
    counter18 = log.create_walker(DataStreamSelector.FromString('counter 16'))

    sg.load_constants()

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 120 seconds')
    sim.run(sg)

    sg.load_constants()

    print([str(x) for x in log._last_values.keys()])

    assert counter15.count() == 2
    assert counter16.count() == 2
    assert counter17.count() == 2
    assert counter18.count() == 2


def test_when_block_clock(parser):
    """Make sure we can compile when blocks (without identifiers)."""

    parser.parse_file(get_path(u'basic_when.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()
    assert sg.user_tick() == 1

    # Now make sure it produces the right output
    counter15 = log.create_walker(DataStreamSelector.FromString('counter 15'))
    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 60 seconds')
    sim.run(sg)
    assert counter15.count() == 0

    sim.step(sg, DataStream.FromString('system input 1025'), 8)
    assert counter15.count() == 1
    counter15.skip_all()

    sim.run(sg)

    assert counter15.count() == 60

    counter15.skip_all()
    sim.step(sg, DataStream.FromString('system input 1026'), 8)
    sim.run(sg)
    assert counter15.count() == 0


def test_when_block_on_event(parser):
    """Make sure on connect and on disconnect work."""

    parser.parse_file(get_path(u'basic_when_on.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()

    sim = SensorGraphSimulator()

    # We should only get a reading in unbuffered 1 on connect and unbuffered 2 on disconnect
    with pytest.raises(StreamEmptyError):
        log.inspect_last(DataStream.FromString('unbuffered 2'))

    sim.step(sg, DataStream.FromString('system input 1025'), 8)
    assert log.inspect_last(DataStream.FromString('unbuffered 1')).value == 0

    with pytest.raises(StreamEmptyError):
        log.inspect_last(DataStream.FromString('unbuffered 2'))

    sim.step(sg, DataStream.FromString('system input 1026'), 8)
    assert log.inspect_last(DataStream.FromString('unbuffered 2')).value == 0


def test_on_block(parser):
    """Make sure on count(stream), on value(stream) and on stream work."""

    parser.parse_file(get_path(u'basic_on.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()

    counter1 = log.create_walker(DataStreamSelector.FromString('counter 1'))
    counter2 = log.create_walker(DataStreamSelector.FromString('counter 2'))
    counter3 = log.create_walker(DataStreamSelector.FromString('counter 3'))

    sim = SensorGraphSimulator()
    sim.step(sg, DataStream.FromString('input 1'), 8)

    assert counter1.count() == 0
    assert counter2.count() == 0
    assert counter3.count() == 0

    for i in range(0, 10):
        sim.step(sg, DataStream.FromString('input 1'), 5)

    assert counter1.count() == 10
    assert counter2.count() == 5
    assert counter3.count() == 5


def test_on_block_dual(parser):
    """Make sure on with two conditions works."""

    parser.parse_file(get_path(u'basic_on_dual.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()

    counter1 = log.create_walker(DataStreamSelector.FromString('counter 1'))

    sim = SensorGraphSimulator()
    sim.step(sg, DataStream.FromString('input 1'), 5)

    assert counter1.count() == 0

    sim.step(sg, DataStream.FromString('input 2'), 1)
    for _i in range(0, 10):
        sim.step(sg, DataStream.FromString('input 1'), 5)

    assert counter1.count() == 10

    sim.step(sg, DataStream.FromString('input 2'), 0)
    for _i in range(0, 10):
        sim.step(sg, DataStream.FromString('input 1'), 5)

    assert counter1.count() == 10

def test_latch_block(parser):
    """Make sure that we can compile and run latch blocks."""

    parser.parse_file(get_path(u'basic_latch.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()
    assert sg.user_tick() == 1

    # Now make sure it produces the right output
    counter15 = log.create_walker(DataStreamSelector.FromString('counter 15'))

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 60 seconds')
    sim.run(sg)
    assert counter15.count() == 0

    sim.step(sg, DataStream.FromString('input 10'), 1)
    counter15.skip_all()

    sim.run(sg)

    assert counter15.count() == 60

    counter15.skip_all()
    sim.step(sg, DataStream.FromString('input 10'), 0)
    sim.run(sg)
    assert counter15.count() == 0


def test_config_block(parser):
    """Make sure config blocks and statement are parsed."""

    parser.parse_file(get_path(u'basic_config.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    assert len(sg.config_database) == 5
    assert sg.get_config(SlotIdentifier.FromString('controller'), 0x2000) == (u'uint32_t', 5)
    assert sg.get_config(SlotIdentifier.FromString('slot 1'), 0x5000) == (u'uint8_t', 10)
    assert sg.get_config(SlotIdentifier.FromString('slot 2'), 0x5100) == (u'int8_t', -10)
    assert sg.get_config(SlotIdentifier.FromString('slot 3'), 0x5200) == (u'uint8_t[]', u'"[10,20,30,40]"')
    assert sg.get_config(SlotIdentifier.FromString('slot 4'), 0x5300) == (u'string', u'"test"')


def test_copy_statement(parser):
    """Make sure we can copy data using copy."""

    parser.parse_file(get_path(u'basic_copy.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()

    output1 = log.create_walker(DataStreamSelector.FromString('output 1'))
    output2 = log.create_walker(DataStreamSelector.FromString('output 2'))
    output3 = log.create_walker(DataStreamSelector.FromString('output 3'))

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 60 seconds')
    sim.run(sg)

    assert output1.count() == 1
    assert output2.count() == 1
    assert output3.count() == 1
    val1 = output1.pop()
    val2 = output2.pop()
    val3 = output3.pop()

    assert val1.value == 0
    assert val2.value == 60
    assert val3.value == 1


def test_copy_constant_statement(parser):
    """Make sure we can copy constant values."""

    parser.parse_file(get_path(u'basic_copy_constant.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log
    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()

    output1 = log.create_walker(DataStreamSelector.FromString('output 1'))

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 10 seconds')
    sim.run(sg)

    assert output1.count() == 1
    assert output1.pop().value == 15

    sim.step(sg, DataStream.FromString('input 1'), 10)
    assert output1.count() == 1
    assert output1.pop().value == 0x10


def test_copy_count_statement(parser):
    """Make sure we can copy data count using copy count."""

    parser.parse_file(get_path(u'count.sgf'))

    model = DeviceModel()
    parser.compile(model=model)

    sg = parser.sensor_graph
    log = sg.sensor_log

    for x in sg.dump_nodes():
        print(x)

    sg.load_constants()

    output = log.create_walker(DataStreamSelector.FromString('output 1'))

    sim = SensorGraphSimulator()
    sim.stop_condition('run_time 3 seconds')
    sim.step(sg, user_connected, 8) # Simulates a connected user
    sim.run(sg)
    print(output);
    assert output.count() == 1
    print(output)



"""Parser for reading sensor graph files."""

from builtins import str

import pyparsing

from .language import get_language, get_statement
from .statements import statement_map
from .scopes import RootScope
from .stream_allocator import StreamAllocator
from iotile.sg import SensorGraph, SensorLog
from iotile.sg.engine import InMemoryStorageEngine
from iotile.core.exceptions import ArgumentError
from iotile.sg.exceptions import SensorGraphSyntaxError, SensorGraphSemanticError


class SensorGraphFileParser(object):
    """A parser that builds a sensor graph object from a text file specification."""

    def __init__(self):
        self._scope_stack = []
        self.statements = []
        self.sensor_graph = None

    def dump_tree(self, statement=None, indent_level=0):
        """Dump the AST for this parsed file.

        Args:
            statement (SensorGraphStatement): the statement to print
                if this function is called recursively.
            indent_level (int): The number of spaces to indent this
                statement.  Used for recursively printing blocks of
                statements.
        Returns:
            str: The AST for this parsed sg file as a nested
                tree with one node per line and blocks indented.
        """

        out = u""

        indent = u" "*indent_level

        if statement is None:
            for root_statement in self.statements:
                out += self.dump_tree(root_statement, indent_level)
        else:
            out += indent + str(statement) + u'\n'

            if len(statement.children) > 0:
                for child in statement.children:
                    out += self.dump_tree(child, indent_level=indent_level+4)

        return out

    def parse_file(self, sg_file):
        """Parse a sensor graph file into an AST describing the file.

        This function builds the statements list for this parser.
        """

        try:
            with open(sg_file, "r") as inf:
                data = inf.read()
        except IOError:
            raise ArgumentError("Could not read sensor graph file", path=sg_file)

        # convert tabs to spaces so our line numbers match correctly
        data = data.replace(u'\t', u'    ')

        lang = get_language()
        result = lang.parseString(data)

        for statement in result:
            parsed = self.parse_statement(statement, orig_contents=data)
            self.statements.append(parsed)

    def compile(self, model):
        """Compile this file into a SensorGraph.

        You must have preivously called parse_file to parse a
        sensor graph file into statements that are then executed
        by this command to build a sensor graph.

        The results are stored in self.sensor_graph and can be
        inspected before running optimization passes.

        Args:
            model (DeviceModel): The device model that we should compile
                this sensor graph for.
        """

        log = SensorLog(InMemoryStorageEngine(model), model)
        self.sensor_graph = SensorGraph(log, model)

        allocator = StreamAllocator(self.sensor_graph, model)

        self._scope_stack = []

        # Create a root scope
        root = RootScope(self.sensor_graph, allocator)
        self._scope_stack.append(root)

        for statement in self.statements:
            statement.execute(self.sensor_graph, self._scope_stack)

        self.sensor_graph.initialize_remaining_constants()

    def parse_statement(self, statement, orig_contents):
        """Parse a statement, possibly called recursively.

        Args:
            statement (int, ParseResult): The pyparsing parse result that
                contains one statement prepended with the match location
            orig_contents (str): The original contents of the file that we're
                parsing in case we need to convert an index into a line, column
                pair.

        Returns:
            SensorGraphStatement: The parsed statement.
        """

        children = []
        is_block = False
        name = statement.getName()

        # Recursively parse all children statements in a block
        # before parsing the block itself.
        # If this is a non-block statement, parse it using the statement
        # parser to figure out what specific statement it is before
        # processing it further.
        # This two step process produces better syntax error messsages
        if name == 'block':
            children_statements = statement[1]
            for child in children_statements:
                parsed = self.parse_statement(child, orig_contents=orig_contents)
                children.append(parsed)

            statement = statement[0]
            name = statement.getName()
            is_block = True
        else:
            stmt_language = get_statement()
            locn = statement['location']
            statement = statement['match']
            statement_string = str(u"".join(statement.asList()))

            # Try to parse this generic statement into an actual statement.
            # Do this here in a separate step so we have good error messages when there
            # is a problem parsing a step.
            try:
                statement = stmt_language.parseString(statement_string)[0]
            except (pyparsing.ParseException, pyparsing.ParseSyntaxException) as exc:
                raise SensorGraphSyntaxError("Error parsing statement in sensor graph file", message=exc.msg, line=pyparsing.line(locn, orig_contents).strip(), line_number=pyparsing.lineno(locn, orig_contents), column=pyparsing.col(locn, orig_contents))
            except SensorGraphSemanticError as exc:
                # Reraise semantic errors with line information
                raise SensorGraphSemanticError(exc.msg, line=pyparsing.line(locn, orig_contents).strip(), line_number=pyparsing.lineno(locn, orig_contents), **exc.params)

            name = statement.getName()

        if name not in statement_map:
            raise ArgumentError("Unknown statement in sensor graph file", parsed_statement=statement, name=name)

        if is_block:
            return statement_map[name](statement, children=children)

        return statement_map[name](statement)



"""Convert count == 1 triggers that don't have other streams as inputs on unbuffered streams to always.
"""

from iotile.core.exceptions import ArgumentError
from iotile.sg.node import TrueTrigger, FalseTrigger, InputTrigger
from iotile.sg import DataStreamSelector, DataStream
import copy


class ConvertCountOneToAlways(object):
    """Run the remove copy latest optimization pass.

    If there is a node whose input is:
    (unbuffered x when count == 1) => X

    This can be turned into:
    (unbuffered x always)

    Since the node will only be checked when there is data and
    if there is data the count will equal 1.

    This works for unbuffered and input type streams.

    Args:
        sensor_graph (SensorGraph): The sensor graph to run
            the optimization pass on
    """

    def __init__(self):
        pass

    def run(self, sensor_graph, model):
        """Run this optimization pass on the sensor graph

        If necessary, information on the device model being targeted
        can be found in the associated model argument.

        Args:
            sensor_graph (SensorGraph): The sensor graph to optimize
            model (DeviceModel): The device model we're using
        """

        # This check can be done if there is 1 input and it is count == 1
        # and the stream type is input or unbuffered

        for node, inputs, outputs in sensor_graph.iterate_bfs():
            if node.num_inputs != 1:
                continue

            input_a, trigger_a = node.inputs[0]
            if input_a.selector.match_type not in [DataStream.InputType, DataStream.UnbufferedType]:
                continue

            if not isinstance(trigger_a, InputTrigger):
                continue

            if trigger_a.comp_string != u'==':
                continue

            if not trigger_a.use_count:
                continue

            if trigger_a.reference != 1:
                continue

            # here we're looking at count input | unbuffered X == 1
            node.inputs[0] = (input_a, TrueTrigger())



"""Call RPC statement."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ...stream import DataStream


@python_2_unicode_compatible
class CallRPCStatement(SensorGraphStatement):
    """Call an RPC on a tile.

    The form of the statement should be
    call (ident|number) on slot_id => stream

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
    """

    def __init__(self, parsed):
        self.rpc_id = parsed[0]
        self.slot_id = parsed[1]

        self.stream = None
        if 'explicit_stream' in parsed:
            self.stream = parsed['explicit_stream'][0]

        super(CallRPCStatement, self).__init__([])

    def __str__(self):
        if self.stream is not None:
            return u'call 0x%X on %s => %s;' % (self.rpc_id, str(self.slot_id), str(self.stream))

        return u'call 0x%X;' % (self.rpc_id, str(self.slot_id), str(self.stream))

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This adds a single node to the sensor graph with the call_rpc function
        as is processing function.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        alloc = parent.allocator

        trigger_stream, trigger_cond = parent.trigger_chain()
        rpc_const = alloc.allocate_stream(DataStream.ConstantType, attach=True)
        rpc_val = (self.slot_id.address << 16) | self.rpc_id

        stream = self.stream
        if stream is None:
            stream = alloc.allocate_stream(DataStream.UnbufferedType)

        sensor_graph.add_node(u"({} {} && {} always) => {} using call_rpc".format(trigger_stream, trigger_cond, rpc_const, stream))
        sensor_graph.add_constant(rpc_const, rpc_val)




import pytest
from iotile.core.exceptions import ArgumentError
from iotile.sg.sim.stop_conditions import TimeBasedStopCondition


def test_time_based_stop():
    """Make sure we can parse and use TimeBasedStopConditions."""

    cond1 = TimeBasedStopCondition.FromString(u'run_time 15 days')
    assert cond1.max_time == 15*24*60*60

    cond1 = TimeBasedStopCondition.FromString('run_time 15 days')
    assert cond1.max_time == 15*24*60*60

    cond1 = TimeBasedStopCondition.FromString(u'run_time 15 hours')
    assert cond1.max_time == 15*60*60

    cond1 = TimeBasedStopCondition.FromString(u'run_time 1 hour')
    assert cond1.max_time == 60*60

    cond1 = TimeBasedStopCondition.FromString(u'run_time 1 minute')
    assert cond1.max_time == 60

    cond1 = TimeBasedStopCondition.FromString(u'run_time 0x10 seconds')
    assert cond1.max_time == 16
    assert not cond1.should_stop(15, 15, None)
    assert cond1.should_stop(17, 17, None)

    with pytest.raises(ArgumentError):
        TimeBasedStopCondition.FromString('run_time random thing')



"""Add a data streamer to the sensor graph."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from iotile.sg.exceptions import SensorGraphSemanticError
from iotile.sg import SlotIdentifier
from iotile.sg.streamer import DataStreamer


@python_2_unicode_compatible
class StreamerStatement(SensorGraphStatement):
    """A streamer definition that adds an output to the sensor graph.

    [manual] [encrypted|signed] [realtime] streamer on <selector> [to <slot>];

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
    """

    def __init__(self, parsed):
        realtime = 'realtime' in parsed
        encrypted = 'security' in parsed and parsed['security'] == 'encrypted'
        signed = 'security' in parsed and parsed['security'] == 'signed'
        self.auto = 'manual' not in parsed

        self.with_other = None
        if 'with_other' in parsed:
            self.with_other = parsed['with_other']
            self.auto = False

        dest = SlotIdentifier.FromString('controller')
        if 'explicit_tile' in parsed:
            dest = parsed['explicit_tile']

        selector = parsed['selector']

        # Make sure all of the combination are valid
        if realtime and (encrypted or signed):
            raise SensorGraphSemanticError("Realtime streamers cannot be either signed or encrypted")

        super(StreamerStatement, self).__init__([])

        self.dest = dest
        self.selector = selector

        if realtime:
            self.report_format = u'individual'
        elif signed:
            self.report_format = u'signedlist_userkey'
        elif encrypted:
            raise SensorGraphSemanticError("Encrypted streamers are not yet supported")
        else:
            self.report_format = u'hashedlist'

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This adds a single DataStraemer to the current sensor graph

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        walker = sensor_graph.sensor_log.create_walker(self.selector)
        streamer = DataStreamer(walker, self.dest, self.report_format, self.auto, with_other=self.with_other)
        sensor_graph.add_streamer(streamer)

    def __str__(self):
        return u'streamer {} (format = {});'.format(self.selector, self.report_format)



"""Base class for all statements supported in sensor graph files."""


class SensorGraphStatement(object):
    """The base class for an AST of statements in a sensor graph file.

    Args:
        children (list(SensorGraphStatement)): A list of children statements
            to this node (optional and initialized to [] if not passed)
    """

    def __init__(self, children=None):
        if children is None:
            children = []

        self.children = children

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This function will likely modify the sensor_graph and will possibly
        also add to or remove from the scope_tree.  If there are children nodes
        they will be called after execute_before and before execute_after,
        allowing block statements to sandwich their children in setup and teardown
        functions.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        self.execute_before(sensor_graph, scope_stack)

        for child in self.children:
            child.execute(sensor_graph, scope_stack)

        self.execute_after(sensor_graph, scope_stack)

    def execute_before(self, sensor_graph, scope_stack):
        """Execute statement before children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        pass

    def execute_after(self, sensor_graph, scope_stack):
        """Execute statement after children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        pass




from .scope import Scope
from ...stream import DataStream
from ...slot import SlotIdentifier
from ...node import InputTrigger
from ...exceptions import SensorGraphSemanticError
from ...known_constants import system_tick, user_tick, config_user_tick_secs


class RootScope(Scope):
    """The global scope that contains all others.

    Args:
        sensor_graph (SensorGraph): The SensorGraph we
            are operating on.
        alloc (StreamAllocator): The global stream allocator we
            are using.
    """

    def __init__(self, sensor_graph, alloc):
        super(RootScope, self).__init__(u"Root Scope", sensor_graph, alloc, None)

        self.system_tick = None
        self.user_tick = None

        self._setup()

    def _setup(self):
        """Prepare for code generation by setting up root clock nodes.

        These nodes are subsequently used as the basis for all clock operations.
        """

        # Create a root system tick and user tick node
        systick = self.allocator.allocate_stream(DataStream.CounterType, attach=True)
        usertick = self.allocator.allocate_stream(DataStream.CounterType, attach=True)

        self.sensor_graph.add_node("({} always) => {} using copy_all_a".format(system_tick, systick))
        self.sensor_graph.add_node("({} always) => {} using copy_all_a".format(user_tick, usertick))
        self.sensor_graph.add_config(SlotIdentifier.FromString('controller'), config_user_tick_secs, 'uint32_t', 1)

        self.system_tick = systick
        self.user_tick = usertick

    def trigger_chain(self):
        """Return a NodeInput tuple for creating a node.

        Returns:
            (StreamIdentifier, InputTrigger)
        """

        raise SensorGraphSemanticError("There is no trigger chain in the root scope since no triggering criteria have been set")

    def clock(self, interval):
        """Return a NodeInput tuple for triggering an event every interval.

        Args:
            interval (int): The interval (in seconds) at which this input should
                trigger.
        """

        if (interval % 10) == 0:
            tick = self.allocator.attach_stream(self.system_tick)
            count = interval // 10
        else:
            tick = self.allocator.attach_stream(self.user_tick)
            count = interval

        trigger = InputTrigger(u'count', '>=', count)
        return (tick, trigger)



"""Globally available processing functions that can be used in sensor graph nodes.

These functions should not be used directly but rather will be invoked by name
when creating a sensor graph node inside the SensorGraph class.  They are found
by looking at the installed python packages using pkg_resources.
"""

from iotile.core.exceptions import HardwareError
from iotile.sg import StreamEmptyError
from iotile.core.hw.reports import IOTileReading


def copy_all_a(input_a, *other_inputs, **kwargs):
    """Copy all readings in input a into the output.

    All other inputs are skipped so that after this function
    runs there are no readings left in any of the input walkers
    when the function finishes, even if it generated no output
    readings.

    Returns:
        list(IOTileReading)
    """

    output = []
    while input_a.count() > 0:
        output.append(input_a.pop())

    for input_x in other_inputs:
        input_x.skip_all()

    return output


def copy_latest_a(input_a, *other_inputs, **kwargs):
    """Copy the latest reading from input a into the output.

    All other inputs are skipped to that after this function
    runs there are no readings left in any of the input walkers
    even if no output is generated.

    Returns:
        list(IOTileReading)
    """

    output = []

    last_reading = None

    if input_a.selector.inexhaustible:
        last_reading = input_a.pop()
    else:
        while input_a.count() > 0:
            last_reading = input_a.pop()

    if last_reading is not None:
        output = [last_reading]

    for input_x in other_inputs:
        input_x.skip_all()

    return output


def copy_count_a(input_a, *other_inputs, **kwargs):
    """Copy the latest reading from input a into the output.

    All other inputs are skipped to that after this function
    runs there are no readings left in any of the input walkers
    even if no output is generated.

    Returns:
        list(IOTileReading)
    """

    output = []

    count = input_a.count()

    input_a.skip_all();

    for input_x in other_inputs:
        input_x.skip_all()

    return [IOTileReading(0, 0, count)]


def call_rpc(*inputs, **kwargs):
    """Call an RPC based on the encoded value read from input b.

    The response of the RPC must be a 4 byte value that is used as
    the output of this call.  The encoded RPC must be a 32 bit value
    encoded as "BBH":
        B: ignored, should be 0
        B: the address of the tile that we should call
        H: The id of the RPC to call

    All other readings are then skipped so that there are no
    readings in any input queue when this function returns

    Returns:
        list(IOTileReading)
    """

    rpc_executor = kwargs['rpc_executor']

    output = []
    try:
        value = inputs[1].pop()

        addr = value.value >> 16
        rpc_id = value.value & 0xFFFF

        reading_value = rpc_executor.rpc(addr, rpc_id)
        output.append(IOTileReading(0, 0, reading_value))
    except (HardwareError, StreamEmptyError):
        pass

    for input_x in inputs:
        input_x.skip_all()

    return output


def trigger_streamer(*inputs, **kwargs):
    """Trigger a streamer based on the index read from input b.

    Returns:
        list(IOTileReading)
    """

    # TODO: This function does nothing currently
    return [IOTIleReading(0, 0, 0)]



"""A simple RPC executor that always responds with zeros."""

import struct
from .executor import RPCExecutor
from iotile.core.hw.hwmanager import HardwareManager


class SemihostedRPCExecutor(RPCExecutor):
    """An RPC executor that runs RPCs on an IOTile device.

    This is used for "semihosting" a sensor graph where the SG engine
    runs on your computer but all RPCs are executed on a separete
    IOTile device.

    Semihosting is not 100% the same as running the SensorGraph on the
    device itself because:

    1. system inputs and inputs are currently not forwarded from the
       device to the computer so you cannot respond to events from tiles.
       (This will be addressed in the future)
    2. config variables set in the sensor graph are currently not set
       in the device, which means tiles that have required configurations need to
       be manually configured before semihosting a sensor graph.
       (This will be addressed in the future)

    Args:
        port (str): The port we should use to create a HardwareManager instance
            conected to our device.
        device_id (int): The device id we should connect to.
    """

    def __init__(self, port, device_id):
        self.hw = HardwareManager(port=port)
        self.hw.connect(device_id)

        super(SemihostedRPCExecutor, self).__init__()

    def _call_rpc(self, address, rpc_id, payload):
        """Call an RPC with the given information and return its response.

        Must raise a hardware error of the appropriate kind if the RPC
        can not be executed correctly.  Otherwise it should return the binary
        response payload received from the RPC.

        Args:
            address (int): The address of the tile we want to call the RPC
                on
            rpc_id (int): The id of the RPC that we want to call
            payload (bytes, bytearray): The data that we want to send as the payload
        """

        # FIXME: Set a timeout of 1.1 seconds to make sure we fail if the device hangs but
        #        this should be long enough to accomodate any actual RPCs we need to send.

        status, response = self.hw.stream.send_rpc(address, rpc_id >> 8, rpc_id & 0xFF, payload, timeout=1.1)
        return response



"""A simple RPC executor that always responds with zeros."""

import struct
from .executor import RPCExecutor


class NullRPCExecutor(RPCExecutor):
    """A mock RPCExecutor that always returns 0."""

    def _call_rpc(self, address, rpc_id, payload):
        """Call an RPC with the given information and return its response.

        Must raise a hardware error of the appropriate kind if the RPC
        can not be executed correctly.  Otherwise it should return the binary
        response payload received from the RPC.

        Args:
            address (int): The address of the tile we want to call the RPC
                on
            rpc_id (int): The id of the RPC that we want to call
            payload (bytes, bytearray): The data that we want to send as the payload
        """

        return struct.pack("<L", 0)



"""Configuration object describing a streamer."""

from builtins import str
from iotile.core.exceptions import ArgumentError


class DataStreamer(object):
    """In charge of streaming data remotely from a device via a communication tile.

    DataStreamers are created to listen for data from a stream, package it using some
    packaging scheme and then attempt to send it remotely when triggered (possibly
    automatically by the presence of data).  They can be configured to keep trying to
    send data until it is accepted by the other side.

    Args:
        walker (StreamWalker): The walker identifying the data that will be streamed
        dest_tile (SlotIdentifier): The destination tile that we will stream data to
        report_format (str): A string constant defined in DataStreamer.KnownFormats
        automatic (bool): Whether the streamer should trigger whenever there is data
            or wait until it is manually triggered
        report_type (str): A string constant defined in DataStreamer.KnownTypes
        with_other (int): The index of another streamer that this streamer should
            use to know when to trigger.  Defaults to None in which case the streamer
            triggers on its own.  The combination of automatic and with_other cannot
            be specified.
    """

    KnownTypes = {u'broadcast': 1, u'telegram': 1 << 1, u'synchronous': 1 << 2}
    KnownFormats = {u'individual': 0, u'hashedlist': 1, u'signedlist_userkey': 2, u'signedlist_devicekey': 3}

    def __init__(self, walker, dest_tile, report_format, automatic, report_type=u'telegram', with_other=None):
        report_format = str(report_format)
        report_type = str(report_type)

        if report_format not in DataStreamer.KnownFormats:
            raise ArgumentError("Unknown report format in DataStreamer constructor", report_format=report_format, known_formats=DataStreamer.KnownFormats.keys())

        if report_type not in DataStreamer.KnownTypes:
            raise ArgumentError("Unknown report type in DataStreamer constructor", report_type=report_type, known_types=DataStreamer.KnownTypes.keys())

        self.walker = walker
        self.dest = dest_tile
        self.format = report_format
        self.automatic = automatic
        self.report_type = report_type
        self.with_other = with_other

    def has_data(self):
        """Check if we have data that could be streamed."""

        return self.walker.count() > 0



"""Blocks that execute when someone is connected."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ...known_constants import user_connected, user_disconnected
from ...node import InputTrigger
from ...stream import DataStream
from ..scopes import ClockScope


@python_2_unicode_compatible
class WhenBlock(SensorGraphStatement):
    """A block of statements that should run when someone is connected to the device.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        children(list(SensorGraphStatement)): The statements that are
            part of this when block.
    """

    def __init__(self, parsed, children):
        self.slot_id = parsed[0]

        super(WhenBlock, self).__init__(children)

    def __str__(self):
        return u"when connected to %s" % (str(self.slot_id),)

    def execute_before(self, sensor_graph, scope_stack):
        """Execute statement before children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        alloc = parent.allocator

        # We want to create a gated clock that only fires when there is a connection
        # to a communication tile.  So we create a latching constant stream that is used to gate the
        # clock passed through from the previous scope.
        connect_stream = alloc.allocate_stream(DataStream.UnbufferedType, attach=True)
        disconnect_stream = alloc.allocate_stream(DataStream.UnbufferedType, attach=True)
        latch_stream = alloc.allocate_stream(DataStream.ConstantType, attach=True)
        latch_on_stream = alloc.allocate_stream(DataStream.ConstantType, attach=True)
        latch_off_stream = alloc.allocate_stream(DataStream.ConstantType, attach=True)
        clock_stream = alloc.allocate_stream(DataStream.CounterType)  # Don't attach because it's not an input anywhere

        parent_clock = parent.clock(1)

        sensor_graph.add_node(u"({} always) => {} using copy_latest_a".format(user_connected, connect_stream))
        sensor_graph.add_node(u"({} always) => {} using copy_latest_a".format(user_disconnected, disconnect_stream))

        sensor_graph.add_node(u"({} always && {} when value=={}) => {} using copy_latest_a".format(latch_on_stream, connect_stream, self.slot_id.address, latch_stream))
        sensor_graph.add_node(u"({} always && {} when value=={}) => {} using copy_latest_a".format(latch_off_stream, disconnect_stream, self.slot_id.address, latch_stream))

        sensor_graph.add_node(u"({} {} && {} when value == 1) => {} using copy_latest_a".format(parent_clock[0], parent_clock[1], latch_stream, clock_stream))

        sensor_graph.add_constant(latch_on_stream, 1)
        sensor_graph.add_constant(latch_off_stream, 0)
        sensor_graph.add_constant(latch_stream, 0)

        new_scope = ClockScope(sensor_graph, scope_stack, (clock_stream, InputTrigger(u'count', u'==', 1)), 1)

        # Add two new identifiers to the scope for supporting on connect and on disconnect events
        new_scope.add_identifier('connect', connect_stream)
        new_scope.add_identifier('disconnect', disconnect_stream)
        scope_stack.append(new_scope)

    def execute_after(self, sensor_graph, scope_stack):
        """Execute statement after children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        scope_stack.pop()



"""Classes for representing and selecting data streams.

Streams are basically FIFOs with a numeric tag to represent their
identity and a series of readings.  There are two key classes:

DataStream: A wrapper around the tag that identifies the stream

DataStreamSelector: An object that selects a specific stream or a
    class of streams.
"""

# For python 2/3 compatibility using future module
from builtins import str
from future.utils import python_2_unicode_compatible, iteritems
from iotile.core.exceptions import ArgumentError


@python_2_unicode_compatible
class DataStream(object):
    """An immutable specifier of a specific data stream

    Args:
        stream_type (int): The type of the stream
        stream_id (int): The stream identifier
        system (bool): Whether this stream is user defined or
            has global, system-wide meaning.
    """

    BufferedType = 0
    UnbufferedType = 1
    ConstantType = 2
    InputType = 3
    CounterType = 4
    OutputType = 5

    StringToType = {
        u'buffered': 0,
        u'unbuffered': 1,
        u'constant': 2,
        u'input': 3,
        u'counter': 4,
        u'output': 5
    }

    TypeToString = {
        0: u'buffered',
        1: u'unbuffered',
        2: u'constant',
        3: u'input',
        4: u'counter',
        5: u'output'
    }

    # These streams are globally important system stream codes that
    # are included by default in all wildcard stream matching.
    KnownBreakStreams = {
        1024: 'device_reboot'
    }

    def __init__(self, stream_type, stream_id, system=False):
        self.stream_type = stream_type
        self.stream_id = stream_id
        self.system = system

    @property
    def buffered(self):
        return self.stream_type == self.BufferedType or self.stream_type == self.OutputType

    @property
    def output(self):
        return self.stream_type == DataStream.OutputType

    @classmethod
    def FromString(cls, string_rep):
        """Create a DataStream from a string representation.

        The format for stream designators when encoded as strings is:
        [system] (buffered|unbuffered|constant|input|count|output) <integer>

        Args:
            string_rep (str): The string representation to turn into a
                DataStream
        """

        rep = str(string_rep)

        parts = rep.split()
        if len(parts) > 3:
            raise ArgumentError("Too many whitespace separated parts of stream designator", input_string=string_rep)
        elif len(parts) == 3 and parts[0] != u'system':
            raise ArgumentError("Too many whitespace separated parts of stream designator", input_string=string_rep)
        elif len(parts) < 2:
            raise ArgumentError("Too few components in stream designator", input_string=string_rep)

        # Now actually parse the string
        if len(parts) == 3:
            system = True
            stream_type = parts[1]
            stream_id = parts[2]
        else:
            system = False
            stream_type = parts[0]
            stream_id = parts[1]

        try:
            stream_id = int(stream_id, 0)
        except ValueError as exc:
            raise ArgumentError("Could not convert stream id to integer", error_string=str(exc), stream_id=stream_id)

        try:
            stream_type = cls.StringToType[stream_type]
        except KeyError:
            raise ArgumentError("Invalid stream type given", stream_type=stream_type, known_types=cls.StringToType.keys())

        return DataStream(stream_type, stream_id, system)

    @classmethod
    def FromEncoded(self, encoded):
        """Create a DataStream from an encoded 16-bit unsigned integer.

        Returns:
            DataStream: The decoded DataStream object
        """

        stream_type = (encoded >> 12) & 0b1111
        stream_system = bool(encoded & (1 << 11))
        stream_id = (encoded & ((1 << 11) - 1))

        return DataStream(stream_type, stream_id, stream_system)

    def encode(self):
        """Encode this stream as a packed 16-bit unsigned integer.

        Returns:
            int: The packed encoded stream
        """

        return (self.stream_type << 12) | (int(self.system) << 11) | self.stream_id

    def __str__(self):
        type_str = self.TypeToString[self.stream_type]

        if self.system:
            return u'system {} {}'.format(type_str, self.stream_id)

        return u'{} {}'.format(type_str, self.stream_id)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, DataStream):
            return NotImplemented

        return self.system == other.system and self.stream_type == other.stream_type and self.stream_id == other.stream_id


@python_2_unicode_compatible
class DataStreamSelector(object):
    """A specifier that matches DataStreams based on their types and ids.

    Args:
        stream_type (int): The type of the stream to match
        stream_id (int): The stream identifier to match.  If None, all
            streams of the given type are matched.
        stream_specifier (int): One of the 4 below match specifiers.
            These control whether user and/or system values are matched.
    """

    MatchSystemOnly = 1
    MatchCombined = 2
    MatchUserOnly = 3
    MatchUserAndBreaks = 4

    ValidSpecifiers = frozenset([MatchSystemOnly, MatchCombined, MatchUserOnly, MatchUserAndBreaks])

    SpecifierStrings = {
        MatchSystemOnly: u'system',
        MatchCombined: u'combined',
        MatchUserOnly: u'user',
        MatchUserAndBreaks: u''
    }

    SpecifierNames = {y: x for x, y in iteritems(SpecifierStrings)}

    SpecifierEncodings = {
        MatchSystemOnly: (1 << 11),
        MatchUserOnly: 0,
        MatchUserAndBreaks: (1 << 15),
        MatchCombined: (1 << 11) | (1 << 15)
    }

    def __init__(self, stream_type, stream_id, stream_specifier):
        self.match_type = stream_type
        self.match_id = stream_id
        self.match_spec = stream_specifier

        if stream_specifier not in DataStreamSelector.ValidSpecifiers:
            raise ArgumentError("Unknown stream selector specifier", specifier=stream_specifier, known_specifiers=DataStreamSelector.ValidSpecifiers)

    @property
    def input(self):
        """Whether this is a root input stream."""

        return self.match_type == DataStream.InputType

    @property
    def inexhaustible(self):
        """Whether this is a constant stream."""

        return self.match_type == DataStream.ConstantType

    @property
    def singular(self):
        """Whether this selector matches only a single stream."""

        return self.match_id is not None

    @property
    def buffered(self):
        return self.match_type == DataStream.BufferedType or self.match_type == DataStream.OutputType

    @property
    def output(self):
        return self.match_type == DataStream.OutputType

    def as_stream(self):
        """Convert this selector to a DataStream.

        This function will only work if this is a singular selector that
        matches exactly one DataStream.
        """

        if not self.singular:
            raise ArgumentError("Attempted to convert a non-singular selector to a data stream, it matches multiple", selector=self)

        return DataStream(self.match_type, self.match_id, self.match_spec == DataStreamSelector.MatchSystemOnly)

    @classmethod
    def FromStream(cls, stream):
        """Create a DataStreamSelector from a DataStream.

        Args:
            stream (DataStream): The data stream that we want to convert.
        """

        if stream.system:
            specifier = DataStreamSelector.MatchSystemOnly
        else:
            specifier = DataStreamSelector.MatchUserOnly

        return DataStreamSelector(stream.stream_type, stream.stream_id, specifier)

    @classmethod
    def FromString(cls, string_rep):
        """Create a DataStreamSelector from a string.

        The format of the string should either be:

        all <type>
        OR
        <type> <id>

        Where type is [system] <stream type>, with <stream type>
        defined as in DataStream

        Args:
            rep (str): The string representation to convert to a DataStreamSelector
        """

        rep = str(string_rep)

        rep = rep.replace(u'node', '')
        rep = rep.replace(u'nodes', '')

        if rep.startswith(u'all'):
            parts = rep.split()

            spec_string = u''

            if len(parts) == 3:
                spec_string = parts[1]
                stream_type = parts[2]
            elif len(parts) == 2:
                stream_type = parts[1]
            else:
                raise ArgumentError("Invalid wildcard stream selector", string_rep=string_rep)

            try:
                # Remove pluralization that can come with e.g. 'all system outputs'
                if stream_type.endswith(u's'):
                    stream_type = stream_type[:-1]

                stream_type = DataStream.StringToType[stream_type]
            except KeyError:
                raise ArgumentError("Invalid stream type given", stream_type=stream_type, known_types=DataStream.StringToType.keys())

            stream_spec = DataStreamSelector.SpecifierNames.get(spec_string, None)
            if stream_spec is None:
                raise ArgumentError("Invalid stream specifier given (should be system, user, combined or blank)", string_rep=string_rep, spec_string=spec_string)

            return DataStreamSelector(stream_type, None, stream_spec)

        # If we're not matching a wildcard stream type, then the match is exactly
        # the same as a DataStream identifier, so use that to match it.

        stream = DataStream.FromString(rep)
        return DataStreamSelector.FromStream(stream)

    def __str__(self):
        type_str = DataStream.TypeToString[self.match_type]

        if self.match_id is not None:
            if self.match_spec == DataStreamSelector.MatchSystemOnly:
                return u'system {} {}'.format(type_str, self.match_id)

            return u'{} {}'.format(type_str, self.match_id)
        else:
            specifier = DataStreamSelector.SpecifierStrings[self.match_spec]

            if specifier != u'':
                specifier += u' '

            return u'all {}{}s'.format(specifier, type_str)

    def matches(self, stream):
        """Check if this selector matches the given stream

        Args:
            stream (DataStream): The stream to check

        Returns:
            bool: True if this selector matches the stream
        """

        if self.match_type != stream.stream_type:
            return False

        if self.match_id is not None:
            return self.match_id == stream.stream_id

        if self.match_spec == DataStreamSelector.MatchUserOnly:
            return not stream.system
        elif self.match_spec == DataStreamSelector.MatchSystemOnly:
            return stream.system
        elif self.match_spec == DataStreamSelector.MatchUserAndBreaks:
            return (not stream.system) or (stream.system and (stream.stream_id in DataStream.KnownBreakStreams))

        # The other case is that match_spec is MatchCombined, which matches everything
        # regardless of system of user flag
        return True

    def encode(self):
        """Encode this stream as a packed 16-bit unsigned integer.

        Returns:
            int: The packed encoded stream
        """

        match_id = self.match_id
        if match_id is None:
            match_id = (1 << 11) - 1

        return (self.match_type << 12) | DataStreamSelector.SpecifierEncodings[self.match_spec] | match_id

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, DataStreamSelector):
            return NotImplemented

        return self.match_spec == other.match_spec and self.match_type == other.match_type and self.match_id == other.match_id



from toposort import toposort_flatten
from .passes import RemoveCopyLatestPass, ConvertCountOneToAlways, ConvertCopyAllToCopyLatest


class SensorGraphOptimizer(object):
    """An optimizer that runs optimization rules on a sensor graph.

    The optimizer keeps track of which rules are disallowed and not
    run and which rules caused the elimination of which nodes.
    """

    def __init__(self):
        self._known_passes = {}

        # Add in our known optimization passes
        self.add_pass('remove-copy', RemoveCopyLatestPass)
        self.add_pass('convert-always', ConvertCountOneToAlways, before=['remove-copy'])
        self.add_pass('downgrade-copyall', ConvertCopyAllToCopyLatest, before=['convert-always'])

    def add_pass(self, name, opt_pass, before=None, after=None):
        """Add an optimization pass to the optimizer.

        Optimization passes have a name that allows them
        to be enabled or disabled by name.  By default all
        optimization passed are enabled and unordered.  You can
        explicitly specify passes by name that this pass must run
        before or after this passs so that they can be properly
        ordered.

        Args:
            name (str): The name of the optimization pass to allow for
                enabling/disabling it by name
            opt_pass (OptimizationPass): The optimization pass class itself
            before (list(str)): A list of the passes that this pass should
                run before.
            after (list(str)): A list of the passes that this pass should
                run after.
        """

        if before is None:
            before = []
        if after is None:
            after = []

        self._known_passes[name] = (opt_pass, before, after)

    def _order_pases(self, passes):
        """Topologically sort optimization passes.

        This ensures that the resulting passes are run in order
        respecting before/after constraints.

        Args:
            passes (iterable): An iterable of pass names that should
                be included in the optimization passes run.
        """

        passes = set(passes)

        pass_deps = {}

        for opt in passes:
            _, before, after = self._known_passes[opt]

            if opt not in pass_deps:
                pass_deps[opt] = set()

            pass_deps[opt].update(*[x for x in after])

            # For passes that we are before, we may need to
            # preemptively add them to the list early
            for other in before:
                if other not in passes:
                    continue

                if other not in pass_deps:
                    pass_deps[other] = set()

                pass_deps[other].add(opt)

        return toposort_flatten(pass_deps)

    def optimize(self, sensor_graph, model):
        """Optimize a sensor graph by running optimization passes.

        The passes are run one at a time and modify the sensor graph
        for future passes.

        Args:
            sensor_graph (SensorGraph): The graph to be optimized
            model (DeviceModel): The device that we are optimizing
                for, that OptimizationPass objects are free to use
                to guide their optimizations.
        """

        passes = self._order_pases(self._known_passes.keys())

        for opt_name in passes:
            rerun = True
            pass_instance = self._known_passes[opt_name][0]()

            while rerun:
                rerun = pass_instance.run(sensor_graph, model=model)



from pyparsing import Optional, Word, Literal, Forward, alphas, nums, Group, Regex, ZeroOrMore, oneOf, restOfLine, dblQuotedString, StringEnd, Empty
from ..stream import DataStream, DataStreamSelector
from ..slot import SlotIdentifier


ident = None
rvalue = None
number = None
comp = None
config_type = None
slot_id = None
stream = None
selector = None
stream_trigger = None
time_interval = None
quoted_string = None
semi = None
comment = None

streamer_stmt = None

callrpc_stmt = None
simple_statement = None
generic_statement = None

block_id = None
block_bnf = None
statement = None
sensor_graph = None


def _create_primitives():
    global ident, rvalue, number, quoted_string, semi, time_interval, slot_id, comp, config_type, stream, comment, stream_trigger, selector

    if ident is not None:
        return

    semi = Literal(u';').suppress()
    ident = Word(alphas+u"_", alphas + nums + u"_")
    number = Regex(u'((0x[a-fA-F0-9]+)|[+-]?[0-9]+)').setParseAction(lambda s, l, t: [int(t[0], 0)])
    quoted_string = dblQuotedString

    comment = Literal('#') + restOfLine

    rvalue = number | quoted_string

    # Convert all time intervals into an integer number of seconds
    time_unit_multipliers = {
        u'second': 1,
        u'seconds': 1,
        u'minute': 60,
        u'minutes': 60,
        u'hour': 60*60,
        u'hours': 60*60,
        u'day': 60*60*24,
        u'days': 60*60*24,
        u'month': 60*60*24*30,
        u'months': 60*60*24*30,
        u'year': 60*60*24*365,
        u'years': 60*60*24*365,
    }

    config_type = oneOf('uint8_t uint16_t uint32_t int8_t int16_t int32_t uint8_t[] uint16_t[] uint32_t[] int8_t[] int16_t[] int32_t[] string')
    comp = oneOf('> < >= <= == ~=')

    time_unit = oneOf(u"second seconds minute minutes hour hours day days week weeks month months year years")
    time_interval = (number + time_unit).setParseAction(lambda s, l, t: [t[0]*time_unit_multipliers[t[1]]])

    slot_id = Literal(u"controller") | (Literal(u'slot') + number)
    slot_id.setParseAction(lambda s,l,t: [SlotIdentifier.FromString(u' '.join([str(x) for x in t]))])

    stream_modifier = Literal("system") | Literal("user") | Literal("combined")

    stream = Optional(Literal("system")) + oneOf("buffered unbuffered input output counter constant") + number + Optional(Literal("node"))
    stream.setParseAction(lambda s,l,t: [DataStream.FromString(u' '.join([str(x) for x in t]))])

    all_selector = Optional(Literal("all")) + Optional(stream_modifier) + oneOf("buffered unbuffered inputs outputs counters constants") + Optional(Literal("nodes"))
    all_selector.setParseAction(lambda s,l,t: [DataStreamSelector.FromString(u' '.join([str(x) for x in t]))])
    one_selector = Optional(Literal("system")) + oneOf("buffered unbuffered input output counter constant") + number + Optional(Literal("node"))
    one_selector.setParseAction(lambda s,l,t: [DataStreamSelector.FromString(u' '.join([str(x) for x in t]))])

    selector = one_selector | all_selector

    trigger_comp = oneOf('> < >= <= ==')
    stream_trigger = Group((Literal(u'count') | Literal(u'value')) + Literal(u'(').suppress() - stream - Literal(u')').suppress() - trigger_comp - number).setResultsName('stream_trigger')


def _create_simple_statements():
    global ident, rvalue, simple_statement, semi, comp, number, slot_id, callrpc_stmt, generic_statement, streamer_stmt, stream, selector

    if simple_statement is not None:
        return

    meta_stmt = Group(Literal('meta').suppress() + ident + Literal('=').suppress() + rvalue + semi).setResultsName('meta_statement')
    require_stmt = Group(Literal('require').suppress() + ident + comp + rvalue + semi).setResultsName('require_statement')
    set_stmt = Group(Literal('set').suppress() + (ident | number) + Literal("to").suppress() + rvalue + Optional(Literal('as').suppress() + config_type) + semi).setResultsName('set_statement')
    callrpc_stmt = Group(Literal("call").suppress() + (ident | number) + Literal("on").suppress() + slot_id + Optional(Literal("=>").suppress() + stream('explicit_stream')) + semi).setResultsName('call_statement')
    streamer_stmt = Group(Optional(Literal("manual")('manual')) + Optional(oneOf(u'encrypted signed')('security')) + Optional(Literal(u'realtime')('realtime')) + Literal('streamer').suppress() -
                          Literal('on').suppress() - selector('selector') - Optional(Literal('to').suppress() - slot_id('explicit_tile')) - Optional(Literal('with').suppress() - Literal('streamer').suppress() - number('with_other')) - semi).setResultsName('streamer_statement')
    copy_stmt = Group(Literal("copy").suppress() - Optional(oneOf("all count average")('modifier')) - Optional(stream('explicit_input') | number('constant_input')) - Literal("=>") - stream("output") - semi).setResultsName('copy_statement')
    trigger_stmt = Group(Literal("trigger") - Literal("streamer") - number('index') - semi).setResultsName('trigger_statement')

    simple_statement = meta_stmt | require_stmt | set_stmt | callrpc_stmt | streamer_stmt | trigger_stmt | copy_stmt

    # In generic statements, keep track of the location where the match started for error handling
    locator = Empty().setParseAction(lambda s, l, t: l)('location')
    generic_statement = Group(locator + Group(ZeroOrMore(Regex(u"[^{};]+")) + Literal(u';'))('match')).setResultsName('unparsed_statement')


def _create_block_bnf():
    global block_bnf, time_interval, slot_id, statement, block_id, ident, stream

    if block_bnf is not None:
        return

    trigger_clause = Group(stream_trigger | Group(stream).setResultsName('stream_always') | Group(ident).setResultsName('identifier'))

    every_block_id = Group(Literal(u'every').suppress() - time_interval).setResultsName('every_block')
    when_block_id = Group(Literal(u'when').suppress() + Literal("connected").suppress() - Literal("to").suppress() - slot_id).setResultsName('when_block')
    latch_block_id = Group(Literal(u'when').suppress() - stream_trigger).setResultsName('latch_block')
    config_block_id = Group(Literal(u'config').suppress() - slot_id).setResultsName('config_block')
    on_block_id = Group(Literal(u'on').suppress() - trigger_clause.setResultsName('triggerA') - Optional((Literal("and") | Literal("or")) - trigger_clause.setResultsName('triggerB'))).setResultsName('on_block')

    block_id = every_block_id | when_block_id | latch_block_id | config_block_id | on_block_id

    block_bnf = Forward()
    statement = generic_statement | block_bnf

    block_bnf << Group(block_id + Group(Literal(u'{').suppress() + ZeroOrMore(statement) + Literal(u'}').suppress())).setResultsName('block')


def get_language():
    """Create or retrieve the parse tree for defining a sensor graph."""

    global sensor_graph, statement

    if sensor_graph is not None:
        return sensor_graph

    _create_primitives()
    _create_simple_statements()
    _create_block_bnf()

    sensor_graph = ZeroOrMore(statement) + StringEnd()
    sensor_graph.ignore(comment)

    return sensor_graph


def get_statement():
    """Create or retrieve the parse tree for a simple statement."""

    global simple_statement

    get_language()

    return simple_statement



"""Sensor graph nodes.

Sensor graph nodes are the basic input/output unit of a sensor graph.
They take a number of inputs, run them through a translation function
and create an output.  The output can then be linked to other nodes
to create a graph structure (hence the name SensorGraph).
"""

from builtins import str
from future.utils import python_2_unicode_compatible
from .exceptions import TooManyInputsError, TooManyOutputsError, ProcessingFunctionError
from iotile.core.exceptions import ArgumentError
from .walker import InvalidStreamWalker
from copy import copy


@python_2_unicode_compatible
class InputTrigger(object):
    """A triggering condition for a graph node input.

    Node inputs can trigger either based on how many readings are
    available to read, or based on the value of the last reading.

    The trigger is evaluated as:

    <source> <comparator> <reference_value>

    Args:
        source (str): either count or value to trigger on either
            a count of available readings or the value of the
            last reading.
        comparator (str): a string describing the comparator.
            Possible options are >, >=, <, =, <=
        reference_value (int): The value to compare source against.
    """

    def __init__(self, source, comparator, reference_value):
        source = str(source)
        if source not in [u'count', u'value']:
            raise ArgumentError("Unknwon source for input trigger, should be count or value", source=source)

        self.use_count = False
        if source == u'count':
            self.use_count = True

        known_comps = {
            u'>': self._gt_comp,
            u'>=': self._ge_comp,
            u'<': self._lt_comp,
            u'<=': self._le_comp,
            u'==': self._eq_comp
        }

        comparator = str(comparator)
        if comparator not in known_comps:
            raise ArgumentError("Unkown comparison function for input trigger", comparator=comparator)

        self.comp_function = known_comps[comparator]
        self.comp_string = comparator
        self.reference = reference_value

    def format_trigger(self, stream):
        """Create a user understandable string like count(stream) >= X.

        Args:
            stream (DataStream): The stream to use to format ourselves.

        Returns:
            str: The formatted string
        """

        src = u'value'
        if self.use_count:
            src = u'count'

        return u"{}({}) {} {}".format(src, stream, self.comp_string, self.reference)

    def triggered(self, walker):
        """Check if this input is triggered on the given stream walker.

        Args:
            walker (StreamWalker): The walker to check

        Returns:
            bool: Whether this trigger is triggered or not
        """

        if self.use_count:
            comp_value = walker.count()
        else:
            if walker.count() == 0:
                return False

            comp_value = walker.peek().value

        return self.comp_function(comp_value, self.reference)

    def _gt_comp(self, comp, ref):
        return comp > ref

    def _ge_comp(self, comp, ref):
        return comp >= ref

    def _lt_comp(self, comp, ref):
        return comp < ref

    def _le_comp(self, comp, ref):
        return comp <= ref

    def _eq_comp(self, comp, ref):
        return comp == ref

    def __str__(self):
        if self.use_count:
            return u"when count {} {}".format(self.comp_string, self.reference)

        return u"when value {} {}".format(self.comp_string, self.reference)


class FalseTrigger(object):
    """Simple trigger that always returns False."""

    def triggered(self, walker):
        return False


@python_2_unicode_compatible
class TrueTrigger(object):
    """Simple trigger that always returns True."""

    def triggered(self, walker):
        return True

    def __str__(self):
        return u"always"

    def format_trigger(self, stream):
        """Create a user understandable string like count(stream) >= X.

        Args:
            stream (DataStream): The stream to use to format ourselves.

        Returns:
            str: The formatted string
        """


        return u"{}".format(stream)


@python_2_unicode_compatible
class SGNode(object):
    """A node in a graph based processing structure.

    Each node has a function that it uses to consume inputs and
    create one output.  That output can be linked to the input
    of other nodes in a graph based structure.

    Args:
        stream (DataStream): The name of the data stream generated
            by this processing node
        model (DeviceModel): The device model that we are building this
            node for so we can constrain the maximum number of inputs
            and outputs.
    """

    AndTriggerCombiner = 0
    OrTriggerCombiner = 1

    def __init__(self, stream, model):
        max_inputs = model.get('max_node_inputs')
        max_outputs = model.get('max_node_outputs')

        self.inputs = [(InvalidStreamWalker(None), FalseTrigger())]*max_inputs
        self.outputs = []
        self.stream = stream
        self.func_name = None
        self.func = None
        self.trigger_combiner = SGNode.OrTriggerCombiner

        self.max_outputs = max_outputs

    def connect_input(self, index, walker, trigger=None):
        """Connect an input to a stream walker.

        If the input is already connected to something an exception is thrown.
        Otherwise the walker is used to read inputs for that input.

        A triggering condition can optionally be passed that will determine
        when this input will be considered as triggered.

        Args:
            index (int): The index of the input that we want to connect
            walker (StreamWalker): The stream walker to use for the input
            trigger (InputTrigger): The trigger to use for the input.  If
                no trigger is specified, the input is considered to always be
                triggered (so TrueTrigger is used)
        """

        if trigger is None:
            trigger = TrueTrigger()

        if index >= len(self.inputs):
            raise TooManyInputsError("Input index exceeded max number of inputs", index=index, max_inputs=len(self.inputs), stream=self.stream)

        self.inputs[index] = (walker, trigger)

    def input_streams(self):
        """Return a list of DataStream objects for all singular input streams.

        This function only returns individual streams, not the streams that would
        be selected from a selector like 'all outputs' for example.

        Returns:
            list(DataStream): A list of all of the individual DataStreams that are inputs
                of the node.  Input selectors that select multiple streams are not included
        """

        streams = []

        for walker, _trigger in self.inputs:
            if walker.selector is None or not walker.selector.singular:
                continue

            streams.append(walker.selector.as_stream())

        return streams

    def find_input(self, stream):
        """Find the input that responds to this stream.

        Args:
            stream (DataStream): The stream to find

        Returns:
            (index, None): The index if found or None
        """

        for i, input_x in enumerate(self.inputs):
            if input_x[0].matches(stream):
                return i

    @property
    def num_inputs(self):
        """Return the number of connected inputs.

        Returns:
            int: The number of connected inputs
        """

        num = 0

        for walker, _ in self.inputs:
            if not isinstance(walker, InvalidStreamWalker):
                num += 1

        return num

    @property
    def num_outputs(self):
        """Return the number of connected outputs.

        Returns:
            int: The number of nodes connected to this node's output
        """

        return len(self.outputs)

    @property
    def free_outputs(self):
        """Return the number of free output connections.

        Returns:
            int: The number of additional outputs that could be connected.
        """

        return self.max_outputs - self.num_outputs

    def connect_output(self, node):
        """Connect another node to our output.

        This downstream node will automatically be triggered when we update
        our output.

        Args:
            node (SGNode): The node that should receive our output
        """

        if len(self.outputs) == self.max_outputs:
            raise TooManyOutputsError("Attempted to connect too many nodes to the output of a node", max_outputs=self.max_outputs, stream=self.stream)

        self.outputs.append(node)

    def triggered(self):
        """Test if we should trigger our operation.

        We test the trigger condition on each of our inputs and then
        combine those triggers using our configured trigger combiner
        to get an overall result for whether this node is triggered.

        Returns:
            bool: True if we should trigger and False otherwise
        """

        trigs = [x[1].triggered(x[0]) for x in self.inputs]

        if self.trigger_combiner == self.OrTriggerCombiner:
            return True in trigs

        return False not in trigs

    def set_func(self, name, func):
        """Set the processing function to use for this node.

        Args:
            name (str): The name of the function to use.  This is
                just stored for reference in case we need to serialize
                the node later.
            func (callable): A function that is called to process inputs
                for this node.  It should have the following signature:
                callable(input1_walker, input2_walker, ...)
                It should return a list of IOTileReadings that are then pushed into
                the node's output stream
        """

        self.func_name = name
        self.func = func

    def process(self, rpc_executor):
        """Run this node's processing function.

        Args:
            rpc_executor (RPCExecutor): An object capable of executing RPCs
                in case we need to do that.

        Returns:
            list(IOTileReading): A list of IOTileReadings with the results of
                the processing function or an empty list if no results were
                produced
        """

        if self.func is None:
            raise ProcessingFunctionError('No processing function set for node', stream=self.stream)

        results = self.func(*[x[0] for x in self.inputs], rpc_executor=rpc_executor)
        if results is None:
            results = []

        return results

    def __str__(self):
        # FIXME: This only works for a maximum of two inputs
        comb = u'||'

        if self.trigger_combiner == self.AndTriggerCombiner:
            comb = u'&&'

        # Handle one input nodes
        if isinstance(self.inputs[1][0], InvalidStreamWalker):
            return u'({} {}) => {} using {}'.format(self.inputs[0][0].selector, self.inputs[0][1], self.stream, self.func_name)

        return u'({} {} {} {} {}) => {} using {}'.format(self.inputs[0][0].selector, self.inputs[0][1], comb,
                                                         self.inputs[1][0].selector, self.inputs[1][1],
                                                         self.stream, self.func_name)



"""Sensor Graph main object."""

from collections import deque
from pkg_resources import iter_entry_points
from .node_descriptor import parse_node_descriptor
from .slot import SlotIdentifier
from .stream import DataStream
from .known_constants import config_user_tick_secs
from .exceptions import NodeConnectionError, ProcessingFunctionError
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IOTileReading


class SensorGraph(object):
    """A graph based data processing engine.

    Args:
        sensor_log (SensorLog): The sensor log where we should store our
            data
        model (DeviceModel): The device model to use for limiting our
            available resources
    """

    def __init__(self, sensor_log, model=None):
        self.roots = []
        self.nodes = []
        self.streamers = []
        self.constant_database = {}
        self.config_database = {}
        self.sensor_log = sensor_log
        self.model = model

    def add_node(self, node_descriptor):
        """Add a node to the sensor graph based on the description given.

        The node_descriptor must follow the sensor graph DSL and describe
        a node whose input nodes already exist.

        Args:
            node_descriptor (str): A description of the node to be added
                including its inputs, triggering conditions, processing function
                and output stream.
        """

        node, inputs, processor = parse_node_descriptor(node_descriptor, self.model)

        in_root = False

        for i, input_data in enumerate(inputs):
            selector, trigger = input_data

            walker = self.sensor_log.create_walker(selector)
            node.connect_input(i, walker, trigger)

            if selector.input and not in_root:
                self.roots.append(node)
                in_root = True  # Make sure we only add to root list once
            else:
                found = False
                for other in self.nodes:
                    if selector.matches(other.stream):
                        other.connect_output(node)
                        found = True

                if not found and selector.buffered:
                    raise NodeConnectionError("Node has input that refers to another node that has not been created yet", node_descriptor=node_descriptor, input_selector=str(selector), input_index=i)

        # Find and load the processing function for this node
        func = self.find_processing_function(processor)
        if func is None:
            raise ProcessingFunctionError("Could not find processing function in installed packages", func_name=processor)

        node.set_func(processor, func)
        self.nodes.append(node)

    def add_config(self, slot, config_id, config_type, value):
        """Add a config variable assignment to this sensor graph.

        Args:
            slot (SlotIdentifier): The slot identifier that this config
                variable is assigned to.
            config_id (int): The 16-bit id of this config_id
            config_type (str): The type of the config variable, currently
                supported are fixed width integer types and strings
            value (str|int): The value to assign to the config variable.
        """

        if slot not in self.config_database:
            self.config_database[slot] = {}

        self.config_database[slot][config_id] = (config_type, value)

    def add_streamer(self, streamer):
        """Add a streamer to this sensor graph.

        Args:
            streamer (DataStreamer): The streamer we want to add
        """

        self.streamers.append(streamer)

    def add_constant(self, stream, value):
        """Store a constant value for use in this sensor graph.

        Constant assignments occur after all sensor graph nodes have been
        allocated since they must be propogated to all appropriate virtual
        stream walkers.

        Args:
            stream (DataStream): The constant stream to assign the value to
            value (int): The value to assign.
        """

        if stream in self.constant_database:
            raise ArgumentError("Attempted to set the same constant twice", stream=stream, old_value=self.constant_database[stream], new_value=value)

        self.constant_database[stream] = value

    def initialize_remaining_constants(self, value=0):
        """Ensure that all constant streams referenced in the sensor graph have a value.

        Constant streams that are automatically created by the compiler are initialized
        as part of the compilation process but it's possible that the user references
        other constant streams but never assigns them an explicit initial value.  This
        function will initialize them all to a default value (0 if not passed) and
        return the streams that were so initialized.

        Args:
            value (int): Optional value to use to initialize all uninitialized constants.
                Defaults to 0 if not passed.

        Returns:
            list(DataStream): A list of all of the constant streams that were not previously
                initialized and were initialized to the given value in this function.
        """

        remaining = []

        for node, _inputs, _outputs in self.iterate_bfs():
            streams = node.input_streams() + [node.stream]

            for stream in streams:
                if stream.stream_type is not DataStream.ConstantType:
                    continue

                if stream not in self.constant_database:
                    self.add_constant(stream, value)
                    remaining.append(stream)

        return remaining

    def load_constants(self):
        """Load all constants into their respective streams.

        All previous calls to add_constant stored a constant value that
        should be associated with virtual stream walkers.  This function
        actually calls push_stream in order to push all of the constant
        values to their walkers.
        """

        for stream, value in self.constant_database.items():
            self.sensor_log.push(stream, IOTileReading(stream.encode(), 0, value))

    def get_config(self, slot, config_id):
        """Get a config variable assignment previously set on this sensor graph.

        Args:
            slot (SlotIdentifier): The slot that we are setting this config variable
                on.
            config_id (int): The 16-bit config variable identifier.

        Returns:
            (str, str|int): Returns a tuple with the type of the config variable and
                the value that is being set.

        Raises:
            ArgumentError: If the config variable is not currently set on the specified
                slot.
        """

        if slot not in self.config_database:
            raise ArgumentError("No config variables have been set on specified slot", slot=slot)

        if config_id not in self.config_database[slot]:
            raise ArgumentError("Config variable has not been set on specified slot", slot=slot, config_id=config_id)

        return self.config_database[slot][config_id]

    def is_output(self, stream):
        """Check if a stream is a sensor graph output.

        Return:
            bool
        """

        for streamer in self.streamers:
            if streamer.walker.matches(stream):
                return True

        return False

    def user_tick(self):
        """Check the config variables to see if there is a user tick.

        Sensor Graph has a built-in 10 second tick that is sent every 10 seconds
        to allow for triggering timed events.  Users can also set up another tick
        at a different, usually faster interval for their own purposes.

        This is done by setting a config variable on the controller with the desired
        tick interval, which is then interpreted by this function.

        The appropriate config_id to use is listed in known_constants.py

        Returns:
            int: 0 if the user tick is disabled, otherwise the number of seconds
                between each tick
        """

        slot = SlotIdentifier.FromString('controller')

        try:
            var = self.get_config(slot, config_user_tick_secs)
            return var[1]
        except ArgumentError:
            return 0

    def process_input(self, stream, value, rpc_executor):
        """Process an input through this sensor graph.

        The tick information in value should be correct and is transfered
        to all results produced by nodes acting on this tick.

        Args:
            stream (DataStream): The stream the input is part of
            value (IOTileReading): The value to process
            rpc_executor (RPCExecutor): An object capable of executing RPCs
                in case we need to do that.
        """

        self.sensor_log.push(stream, value)

        to_check = deque([x for x in self.roots])

        while len(to_check) > 0:
            node = to_check.popleft()
            if node.triggered():
                results = node.process(rpc_executor)
                for result in results:
                    result.raw_time = value.raw_time
                    self.sensor_log.push(node.stream, result)

                # If we generated any outputs, notify our downstream nodes
                # so that they are also checked to see if they should run.
                if len(results) > 0:
                    to_check.extend(node.outputs)

    def iterate_bfs(self):
        """Generator that yields node, [inputs], [outputs] in breadth first order.

        This generator will iterate over all nodes in the sensor graph, yielding
        a 3 tuple for each node with a list of all of the nodes connected to its
        inputs and all of the nodes connected to its output.

        Returns:
            (SGNode, list(SGNode), list(SGNode)): A tuple for each node in the graph
        """

        working_set = deque(self.roots)
        seen = []

        while len(working_set) > 0:
            curr = working_set.popleft()

            # Now build input and output node lists for this node
            inputs = []
            for walker, _ in curr.inputs:
                for other in seen:
                    if walker.matches(other.stream):
                        inputs.append(other)

            outputs = [x for x in curr.outputs]
            yield curr, inputs, outputs

            working_set.extend(curr.outputs)
            seen.append(curr)

    @classmethod
    def find_processing_function(cls, name):
        """Find a processing function by name.

        This function searches through installed processing functions
        using pkg_resources.

        Args:
            name (str): The name of the function we're looking for

        Returns:
            callable: The processing function
        """

        for entry in iter_entry_points(u'iotile.sg_processor', name):
            return entry.load()

    def dump_nodes(self):
        """Dump all of the nodes in this sensor graph as a list of strings."""

        return [str(x) for x in self.nodes]



"""Command line script to load and run a sensor graph."""

import sys
import argparse
from builtins import str
from iotile.core.exceptions import ArgumentError, IOTileException
from iotile.sg import DeviceModel, DataStreamSelector, SlotIdentifier
from iotile.sg.sim import SensorGraphSimulator
from iotile.sg.sim.hosted_executor import SemihostedRPCExecutor
from iotile.sg.parser import SensorGraphFileParser
from iotile.sg.known_constants import user_connected
from iotile.sg.optimizer import SensorGraphOptimizer


def build_args():
    """Create command line argument parser."""

    parser = argparse.ArgumentParser(description=u'Load and run a sensor graph, either in a simulator or on a physical device.')
    parser.add_argument(u'sensor_graph', type=str, help=u"The sensor graph file to load and run.")
    parser.add_argument(u'--stop', u'-s', action=u"append", default=[], type=str, help=u"A stop condition for when the simulation should end.")
    parser.add_argument(u'--realtime', u'-r', action=u"store_true", help=u"A stop condition for when the simulation should end.")
    parser.add_argument(u'--watch', u'-w', action=u"append", default=[], help=u"A stream to watch and print whenever writes are made.")
    parser.add_argument(u'--disable-optimizer', action="store_true", help=u"disable the sensor graph optimizer completely")
    parser.add_argument(u"--mock-rpc", u"-m", action=u"append", type=str, default=[], help=u"mock an rpc, format should be <slot id>:<rpc_id> = value.  For example -m \"slot 1:0x500a = 10\"")
    parser.add_argument(u"--port", u"-p", help=u"The port to use to connect to a device if we are semihosting")
    parser.add_argument(u"--semihost-device", u"-d", type=lambda x: int(x, 0), help=u"The device id of the device we should semihost this sensor graph on.")
    parser.add_argument(u"-c", u"--connected", action="store_true", help=u"Simulate with a user connected to the device (to enable realtime outputs)")
    return parser


def process_mock_rpc(input_string):
    """Process a mock RPC argument.

    Args:
        input_string (str): The input string that should be in the format
            <slot id>:<rpc id> = value
    """

    spec, equals, value = input_string.partition(u'=')

    if len(equals) == 0:
        print("Could not parse mock RPC argument: {}".format(input_string))
        sys.exit(1)

    try:
        value = int(value.strip(), 0)
    except ValueError as exc:
        print("Could not parse mock RPC value: {}".format(str(exc)))
        sys.exit(1)

    slot, part, rpc_id = spec.partition(u":")
    if len(part) == 0:
        print("Could not parse mock RPC slot/rpc definition: {}".format(spec))
        sys.exit(1)

    try:
        slot = SlotIdentifier.FromString(slot)
    except ArgumentError as exc:
        print("Could not parse slot id in mock RPC definition: {}".format(exc.msg))
        sys.exit(1)

    try:
        rpc_id = int(rpc_id, 0)
    except ValueError as exc:
        print("Could not parse mock RPC number: {}".format(str(exc)))
        sys.exit(1)

    return slot, rpc_id, value

def watch_printer(watch, value):
    """Print a watched value.

    Args:
        watch (DataStream): The stream that was watched
        value (IOTileReading): The value to was seen
    """

    print("({: 8} s) {}: {}".format(value.raw_time, watch, value.value))


def main():
    try:
        executor = None
        parser = build_args()
        args = parser.parse_args()

        model = DeviceModel()

        parser = SensorGraphFileParser()
        parser.parse_file(args.sensor_graph)
        parser.compile(model)

        if not args.disable_optimizer:
            opt = SensorGraphOptimizer()
            opt.optimize(parser.sensor_graph, model=model)

        graph = parser.sensor_graph
        sim = SensorGraphSimulator()

        for stop in args.stop:
            sim.stop_condition(stop)

        for watch in args.watch:
            watch_sel = DataStreamSelector.FromString(watch)
            graph.sensor_log.watch(watch_sel, watch_printer)

        # If we are semihosting, create the appropriate executor connected to the device
        if args.semihost_device is not None:
            executor = SemihostedRPCExecutor(args.port, args.semihost_device)
            sim.rpc_executor = executor

        for mock in args.mock_rpc:
            slot, rpc_id, value = process_mock_rpc(mock)
            sim.rpc_executor.mock(slot, rpc_id, value)

        graph.load_constants()

        try:
            if args.connected:
                sim.step(graph, user_connected, 8)

            sim.run(graph, accelerated=not args.realtime)
        except KeyboardInterrupt:
            pass
    #except IOTileException as exc:
    #    print(str(exc))
    #    return 1
    finally:
        if executor is not None:
            executor.hw.close()



"""Blocks that gate child execution on a condition."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ...node import InputTrigger
from ...stream import DataStream
from ..scopes import ClockScope


@python_2_unicode_compatible
class LatchBlock(SensorGraphStatement):
    """A block of statements that should run when a latch is true.

    The when block is a specific kind of latch block that gates its statements
    on when someone is connected.  The Latch block more generally lets you gate
    on any stream trigger condition.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        children(list(SensorGraphStatement)): The statements that are
            part of this when block.
    """

    def __init__(self, parsed, children):
        cond = parsed[0]
        trigger_type = cond[0]
        stream = cond[1]
        oper = cond[2]
        ref = cond[3]

        trigger = InputTrigger(trigger_type, oper, ref)
        self.stream = stream
        self.trigger = trigger

        super(LatchBlock, self).__init__(children)

    def __str__(self):
        return u"when " + self.trigger.format_trigger(self.stream)

    def execute_before(self, sensor_graph, scope_stack):
        """Execute statement before children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        alloc = parent.allocator

        # We want to create a gated clock that only fires when the latching constant is true
        clock_stream = alloc.allocate_stream(DataStream.CounterType)  # Don't attach because it's not an input anywhere

        parent_clock = parent.clock(1)

        sensor_graph.add_node(u"({} {} && {} {}) => {} using copy_latest_a".format(parent_clock[0], parent_clock[1], self.stream, self.trigger, clock_stream))
        sensor_graph.add_constant(self.stream, 0)

        new_scope = ClockScope(sensor_graph, scope_stack, (clock_stream, InputTrigger(u'count', u'==', 1)), 1)
        scope_stack.append(new_scope)

    def execute_after(self, sensor_graph, scope_stack):
        """Execute statement after children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        scope_stack.pop()



"""A SensorGraph simulator that can drive the sensor graph either in realtime or as fast as possible."""

import time
from monotonic import monotonic
from ..known_constants import system_tick, user_tick, battery_voltage
from .null_executor import NullRPCExecutor
from .stop_conditions import TimeBasedStopCondition
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IOTileReading

class SensorGraphSimulator(object):
    """A simulator for sensor graphs.

    At its core the simulator just sends timed inputs to the sensor graph
    every tick and then sits back and lets the nodes process the data in
    whatever way they are configure.

    There are a few things that need to be configured for the simulator to
    function properly however:

    1. There needs to be a stop condition.  When should the simulation
       stop running?  These conditions can either be simulated times like
       one day or other conditions like a certain number of readings.

    2. There needs to be a way to send RPCs.  Most interesting things inside
       of a sensor graph happen because it sent an RPC and processed its response.
       For this to work those RPCs need to be routed somewhere.  By default, all
       RPCs are accepted and a value of 0 is returned, however, you can configure
       the simulator to execute those RPCs on an actual device if you want.

       This results in a semi-hosted sensor graph, where the sensor graph is running
       on your computer but the rpcs are executed on another IOTile Device, which
       can be very useful for exploration and debugging.
    """

    def __init__(self):
        self.voltage = 3.6
        self.stop_conditions = []
        self._known_conditions = []
        self.watch_streams = []
        self.tick_count = 0
        self._start_tick = 0  # the tick on which the current simulation started
        self.rpc_executor = NullRPCExecutor()

        # Register known stop conditions
        self._known_conditions.append(TimeBasedStopCondition)

    def step(self, sensor_graph, input_stream, value):
        """Step the sensor graph through one since input.

        The internal tick count is not advanced so this function may
        be called as many times as desired to input specific conditions
        without simulation time passing.

        Args:
            sensor_graph (SensorGraph): The sensor graph to run
            input_stream (DataStream): The input stream to push the
                value into
            value (int): The reading value to push as an integer
        """

        reading = IOTileReading(input_stream.encode(), self.tick_count, value)
        sensor_graph.process_input(input_stream, reading, self.rpc_executor)

    def run(self, sensor_graph, include_reset=True, accelerated=True):
        """Run this sensor graph until a stop condition is hit.

        Multiple calls to this function are useful only if
        there has been some change in the stop conditions that would
        cause the second call to not exit immediately.

        Args:
            sensor_graph (SensorGraph): The sensor graph to run
            include_reset (bool): Start the sensor graph run with
                a reset event to match what would happen when an
                actual device powers on.
            accelerated (bool): Whether to run this sensor graph as
                fast as possible or to delay tick events to simulate
                the actual passage of wall clock time.
        """

        self._start_tick = self.tick_count

        if self._check_stop_conditions(sensor_graph):
            return

        if include_reset:
            pass  # TODO: include a reset event here

        # See if there's a user tick that's set
        user_interval = sensor_graph.user_tick()

        while not self._check_stop_conditions(sensor_graph):
            # Process one more one second tick
            now = monotonic()
            next_tick = now + 1.0

            # To match what is done in actual hardware, we incremeent tick count so the first tick
            # is 1.
            self.tick_count += 1

            if user_interval != 0 and (self.tick_count % user_interval) == 0:
                reading = IOTileReading(self.tick_count, user_tick.encode(), self.tick_count)
                sensor_graph.process_input(user_tick, reading, self.rpc_executor)

            if (self.tick_count % 10) == 0:
                reading = IOTileReading(self.tick_count, system_tick.encode(), self.tick_count)
                sensor_graph.process_input(system_tick, reading, self.rpc_executor)

                # Every 10 seconds the battery voltage is reported in 16.16 fixed point format in volts
                reading = IOTileReading(self.tick_count, battery_voltage.encode(), int(self.voltage * 65536))
                sensor_graph.process_input(battery_voltage, reading, self.rpc_executor)

            now = monotonic()

            # If we are trying to execute this sensor graph in realtime, wait for
            # the remaining slice of this tick.
            if (not accelerated) and (now < next_tick):
                time.sleep(next_tick - now)

    def _check_stop_conditions(self, sensor_graph):
        """Check if any of our stop conditions are met.

        Args:
            sensor_graph (SensorGraph): The sensor graph we are currently simulating

        Returns:
            bool: True if we should stop the simulation
        """

        for stop in self.stop_conditions:
            if stop.should_stop(self.tick_count, self.tick_count - self._start_tick, sensor_graph):
                return True

        return False

    def stop_condition(self, condition):
        """Add a stop condition to this simulation.

        Stop conditions are specified as strings and parsed into
        the appropriate internal structures.

        Args:
            condition (str): a string description of the stop condition
        """

        # Try to parse this into a stop condition with each of our registered
        # condition types
        for cond_format in self._known_conditions:
            try:
                cond = cond_format.FromString(condition)
                self.stop_conditions.append(cond)
                return
            except ArgumentError:
                continue

        raise ArgumentError("Stop condition could not be processed by any known StopCondition type", condition=condition, suggestion="It may be mistyped or otherwise invalid.")



"""A grammer for specifying sensor graph nodes."""

from builtins import str
from pyparsing import Word, Regex, nums, hexnums, Literal, Optional, Group, oneOf, QuotedString, ParseException
from .node import SGNode, InputTrigger
from .stream import DataStream, DataStreamSelector

number = Regex('((0x[a-fA-F0-9]+)|[0-9]+)')
combiner = (Literal('&&') | Literal('||'))
symbol = Regex('[a-zA-Z][a-zA-Z_]*')

stream_type = Optional(Literal('system')) + (Literal('input') | Literal('output') | Literal('buffered') | Literal("unbuffered") | Literal("constant") | Literal("counter")) + Optional(Literal("node").suppress())
stream = stream_type + number

trigger_type = (Literal('value') | Literal('count'))
trigger_op = oneOf('> < >= <= ==')

trigger = Literal('always') | (Literal('when').suppress() + trigger_type('type') + trigger_op('op') + number('reference'))

inputstring = stream('input_stream') + trigger

inputdesc2 = Literal('(').suppress() + inputstring('input_a') + combiner('combiner') + inputstring('input_b') + Literal(')').suppress()
inputdesc1 = Literal('(').suppress() + inputstring('input_a') + Literal(')').suppress()

inputdesc = inputdesc1('input1') | inputdesc2('input2')
graph_node = inputdesc + Literal('=>').suppress() + stream('node') + Literal('using').suppress() + symbol('processor')


def parse_node_descriptor(desc, model):
    """Parse a string node descriptor.

    The function creates an SGNode object without connecting its inputs and outputs
    and returns a 3-tuple:

    SGNode, [(input X, trigger X)], <processing function name>

    Args:
        desc (str): A description of the node to be created.
        model (str): A device model for the node to be created that sets any
            device specific limits on how the node is set up.
    """

    try:
        data = graph_node.parseString(desc)
    except ParseException as exc:
        raise  # TODO: Fix this to properly encapsulate the parse error

    stream_desc = u' '.join(data['node'])

    stream = DataStream.FromString(stream_desc)
    node = SGNode(stream, model)

    inputs = []

    if 'input_a' in data:
        input_a = data['input_a']
        stream_a = DataStreamSelector.FromString(u' '.join(input_a['input_stream']))

        trigger_a = None
        if 'type' in input_a:
            trigger_a = InputTrigger(input_a['type'], input_a['op'], int(input_a['reference'], 0))

        inputs.append((stream_a, trigger_a))

    if 'input_b' in data:
        input_a = data['input_b']
        stream_a = DataStreamSelector.FromString(u' '.join(input_a['input_stream']))

        trigger_a = None
        if 'type' in input_a:
            trigger_a = InputTrigger(input_a['type'], input_a['op'], int(input_a['reference'], 0))

        inputs.append((stream_a, trigger_a))

    if 'combiner' in data and str(data['combiner']) == u'&&':
        node.trigger_combiner = SGNode.AndTriggerCombiner
    else:
        node.trigger_combiner = SGNode.OrTriggerCombiner

    processing = data['processor']
    return node, inputs, processing



"""Wellknown streams that have global significance for all IOTile devices."""

from .stream import DataStream

# System streams
# These are streams that are generated automatically by an IOTile Device
# according to internal signals and events.  Some indicate abnormal conditions
# and others are standard signals.

# Battery voltage is sampled and reported every 10 seconds
battery_voltage = DataStream.FromString('system input 0')

# The system tick is fired every 10 seconds
system_tick = DataStream.FromString('system input 2')

# The user tick can be configured by the user to any interval
# 1 second or longer.
user_tick = DataStream.FromString('system input 3')

# These streams receive the address of the tile that an external
# user connected to or disconnected from.
user_connected = DataStream.FromString('system input 1025')
user_disconnected = DataStream.FromString('system input 1026')


# Known config variable ids on the controller

## An optional config variable that sets the time interval in seconds
## for the user tick input to sensor graph
config_user_tick_secs = 0x2000
